//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <fcntl.h>
#include <io.h>
#include <sys\stat.h>
#include "TDirectDraw.h"
//---------------------------------------------------------------------------
TDirectDraw*	DirectDraw = NULL;
//---------------------------------------------------------------------------
//#define	DIRECTX_MOUSE_CURSOR
//---------------------------------------------------------------------------
static inline TDirectDraw *ValidCtrCheck()
{
	return new TDirectDraw(NULL);
}
//---------------------------------------------------------------------------
namespace Tdirectdraw
{
	void __fastcall Register()
	{
		TComponentClass classes[1] = {__classid(TDirectDraw)};
		RegisterComponents("DirectDraw", classes, 0);
	}
}
//---------------------------------------------------------------------------
__fastcall TDirectDraw::TDirectDraw(TComponent* Owner)
	: TComponent(Owner)
{
// create things used for debugging
fErrorIndent = 0;
fErrorValue = 0;
fErrorCount = 0;
fErrorStrings = new TStringList();
fErrorStrings->Sorted = false;
fErrorStrings->Duplicates = dupAccept;
fDebugging = true;

// initialize some values
fDirectDraw = NULL;
fPrimarySurface = NULL;
fBackBuffer = NULL;

fCanvasDC = NULL;
fBackBufferDC = NULL;
fBackBufferDCCnt = 0;
fPrimarySurfaceDC = NULL;
fPrimarySurfaceDCCnt = 0;

fDrawingOnPrimarySurface = false;
fDrawingOnBackBuffer = false;
fAutoFlip = false;

fWidth = Screen->Width;
fHeight = Screen->Height;
fBitsPerPixel = GetDeviceCaps( ((TForm*)Owner)->Canvas->Handle, BITSPIXEL );
fBackBufferCount = 1;

fExclusiveModeSet = false;

// if we have created our DirectDraw object...
if (CreateDirectDraw())
 	{
	// make a note of a successful result
	Debug(DBG_NOTE,"TDirectDraw::","DirectDraw exists");

	// try setting exclusive mode
    if (SetExclusiveMode())
    	{
	    // make a note of a successful result
		Debug(DBG_NOTE,"TDirectDraw::","exclusive mode has been set");
        }
    else
    	Debug(DBG_NOTE,"TDirectDraw::","exclusive mode could not be set");

    // create required display surfaces
    if (CreateSurfaces())
    	{
	    // make a note of a successful result
		Debug(DBG_NOTE,"TDirectDraw::","surfaces have been created");
        }
    else
    	Debug(DBG_NOTE,"TDirectDraw::","surfaces could not be created");

    // let people know
	Debug(DBG_NOTE,"TDirectDraw::","TDirectDraw interface created successfully");
    }
// otherwise display an error messsage
else
	{
    ShowMessage("ERROR>> TDirectDraw >> Could not create the DirectDraw object");
    }
fDebugging = false;

// mouse cursor stuffs
#ifdef DIRECTX_MOUSE_CURSOR
fMouseWidth = GetSystemMetrics(SM_CXCURSOR);
fMouseHeight = GetSystemMetrics(SM_CYCURSOR);
GetCursorPos( &fMousePos );

fMouseCursor = Screen->Cursors[Screen->Cursor];
fMouseUnder = NULL;
fMouseUnderCursor = new Graphics::TBitmap();
fMouseBitmap = new Graphics::TBitmap();

fMouseHasMoved = false;
fMouseFirstTimeDrawn = true;
fMouseBusy = NULL;

fMouseNewRect.Top = fMousePos.y;
fMouseNewRect.Left = fMousePos.x;
fMouseNewRect.Right = fMouseNewRect.Left + fMouseWidth;
fMouseNewRect.Bottom = fMouseNewRect.Top + fMouseHeight;

fMouseOldRect = fMouseNewRect;

// make sure the mouse does not exist
ShowCursor( false );
#endif
}
//---------------------------------------------------------------------------
__fastcall TDirectDraw::~TDirectDraw()
{
// enable the mouse cursor
//Debug(DBG_IN,"TDirectDraw::Destructor()","");

// clean up mouse cursor stuffs
#ifdef DIRECTX_MOUSE_CURSOR
ShowCursor( true );
fMouseUnderCursor->Handle = NULL;
if (fMouseUnder!=NULL) DeleteObject( fMouseUnder );
delete fMouseBitmap;
delete fMouseUnderCursor;
#endif

// if we previously created our DirectDraw object...
if (fDirectDraw != NULL)
	{
	// make a note of a successful result
//	Debug(DBG_NOTE,"TDirectDraw::Destructor()","DirectDraw object was created, time to release");

	// make sure the display is pointing to the GDI surface
    FlipToGDI();

    // release the display surfaces
    ReleaseSurfaces();

    // restore the original video mode
    RestoreDisplayMode();

	// restore normal mode
    SetNormalMode();

    // release the directdraw object
    ReleaseDirectDraw();
    }
// otherwise display an error messsage
else
    ShowMessage("ERROR>> TDirectDraw::Constructor() >> Could not release the DirectDraw object");

// release things used for debugging
if (fErrorStrings != NULL) { fErrorStrings->Clear(); delete fErrorStrings; }
}
//---------------------------------------------------------------------------
// Func: CreateDirectDraw()
// Desc: simply creates a single instance of a IDirectDraw object
bool	__fastcall	TDirectDraw::CreateDirectDraw()
{
fErrorValue = DirectDrawCreate(NULL, &fDirectDraw, NULL);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateDirectDraw()","DirectDraw does not exist");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::ReleaseDirectDraw()
{
if (fDirectDraw == NULL) return false;
// simply call the Release function & use result
fErrorValue = fDirectDraw->Release();
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::ReleaseDirectDraw()","DirectDraw object could not be released");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::SetBackBuffers( int numbackbuffers )
{
if (fDirectDraw == NULL) return;
if (numbackbuffers != fBackBufferCount)
	{
    ReleaseSurfaces();
    fBackBufferCount = numbackbuffers;
    CreateSurfaces();
    }
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SetExclusiveMode()
{
if (fDirectDraw == NULL) return false;
// simply call the SetCooperativeLevel function & use result
fErrorValue = fDirectDraw->SetCooperativeLevel( Application->Handle, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::SetExclusiveMode()","exclusive mode could not be set");
fExclusiveModeSet = (fErrorValue==DD_OK);
return fExclusiveModeSet;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SetNormalMode()
{
if (fDirectDraw == NULL) return false;
// simply call the SetCooperativeLevel function & use result
fErrorValue = fDirectDraw->SetCooperativeLevel( Application->Handle, DDSCL_NORMAL );
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::SetNormalMode()","normal mode could not be set");
if (fErrorValue==DD_OK) fExclusiveModeSet = false;
return !fExclusiveModeSet;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::RestoreDisplayMode()
{
if (fDirectDraw == NULL) return false;
// simply call the RestoreDisplayMode function & use result
fErrorValue = fDirectDraw->RestoreDisplayMode();
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::RestoreVideoMode()","video mode could not be restored");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::CreateSurfaces()
{
if (fDirectDraw == NULL) return false;

// create the primary surface if it's not already created
fErrorValue = DD_OK;
if (fPrimarySurface == NULL)
	{
	// define the properties of the primary surface to be created
	DDSURFACEDESC	ddSurfaceDesc;
	memset(&ddSurfaceDesc,0,sizeof(ddSurfaceDesc));

	ddSurfaceDesc.dwSize = sizeof(ddSurfaceDesc);
	ddSurfaceDesc.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddSurfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
	ddSurfaceDesc.dwBackBufferCount = fBackBufferCount;

	// create the primary surface
	fErrorValue = fDirectDraw->CreateSurface(&ddSurfaceDesc,&fPrimarySurface,NULL);
	if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateSurfaces()","could not create the primary surface");
    else
    	{
        // note a successful result
//        Debug(DBG_NOTE,"TDirectDraw::CreateSurfaces()","the primary surface has been created");

        // only retrieve back buffer if any exist
        if (fBackBufferCount>0)
        	{
			// setup the surface caps ready to retrieve the back buffer
			DDSCAPS			ddsCaps;
			ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;

	        // retrieve a pointer to the 1st back buffer in the list (i theorize)
			fErrorValue = fPrimarySurface->GetAttachedSurface(&ddsCaps,&fBackBuffer);
			if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateSurfaces()","could not retrieve back buffer");
//		    else Debug(DBG_NOTE,"TDirectDraw::CreateSurfaces()","back buffer has been retrieved");
            }
        }
	}
//else
//	Debug(DBG_WARNING,"TDirectDraw::CreateSurfaces()","primary surface has already been created");

return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::ReleaseSurfaces()
{
if (fDirectDraw == NULL) return false;
/*
NOTE:	the back buffer cannot be released directly - all associated back
		buffers are released when you release the primary surface			*/

// get rid of our primary surface & any attached back buffers if they exist
fErrorValue = DD_OK;
if (fPrimarySurface != NULL)
	{
    fErrorValue = fPrimarySurface->Release();
    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::ReleaseSurfaces()","could not release primary surface");
	else
    	{
        fPrimarySurface = NULL;
    	fBackBuffer = NULL;
        }
    }
//else
//	Debug(DBG_WARNING,"TDirectDraw::ReleaseSurfaces()","Primary surface was not allocated, thus no buffers were released");
return (fErrorValue != DD_OK);
}
//---------------------------------------------------------------------------
TCanvas*	__fastcall	TDirectDraw::BeginPrimarySurfaceDrawing(int left, int top, int right, int bottom)
{
if (fDrawingOnPrimarySurface || fDrawingOnBackBuffer) return NULL;

TCanvas* canvas = new TCanvas();
if (canvas != NULL)
	{
    canvas->Handle = GetPrimarySurfaceDC();
	if (canvas->Handle == NULL)
		{
	    ReleasePrimarySurfaceDC();
        delete canvas;
        return NULL;
        }
    }

SelectClipRgn( canvas->Handle, NULL );
HRGN hrgn = CreateRectRgn( left, top, right, bottom);
SelectClipRgn( canvas->Handle, hrgn );
DeleteObject( hrgn );

fDrawingOnPrimarySurface = true;
return canvas;
}
//---------------------------------------------------------------------------
TCanvas*	__fastcall	TDirectDraw::BeginBackBufferDrawing(int left, int top, int right, int bottom, bool autoflip)
{
if (fDrawingOnPrimarySurface || fDrawingOnBackBuffer) return NULL;

TCanvas* canvas = new TCanvas();
if (canvas != NULL)
	{
    canvas->Handle = GetBackBufferDC();
	if (canvas->Handle == NULL)
		{
	    ReleaseBackBufferDC();
        delete canvas;
        return NULL;
        }
    }

fAutoFlip = autoflip;
fAutoFlipClipRect = Rect( left, top, right, bottom );

SelectClipRgn( canvas->Handle, NULL );
HRGN hrgn = CreateRectRgn( left, top, right, bottom);
SelectClipRgn( canvas->Handle, hrgn );
DeleteObject( hrgn );

fDrawingOnBackBuffer = true;
return canvas;
}
//---------------------------------------------------------------------------
void		__fastcall	TDirectDraw::EndPrimarySurfaceDrawing( TCanvas* canvas )
{
if (!fDrawingOnPrimarySurface) return;
if (canvas == NULL) return;
if (canvas->Handle == NULL) return;
if (canvas->Handle != fPrimarySurfaceDC) return;

//SelectClipRgn( canvas->Handle, NULL );

canvas->Handle = ReleasePrimarySurfaceDC();
delete canvas;

fDrawingOnPrimarySurface = false;
}
//---------------------------------------------------------------------------
void		__fastcall	TDirectDraw::EndBackBufferDrawing( TCanvas* canvas )
{
if (!fDrawingOnBackBuffer) return;
if (canvas == NULL) return;
if (canvas->Handle == NULL) return;
if (canvas->Handle != fBackBufferDC) return;

//SelectClipRgn( canvas->Handle, NULL );

canvas->Handle = ReleaseBackBufferDC();
delete canvas;

fDrawingOnBackBuffer = false;

if (fAutoFlip)
	{
    FlipBackBufferRect( fAutoFlipClipRect.Left, fAutoFlipClipRect.Top,
    					fAutoFlipClipRect.Right, fAutoFlipClipRect.Bottom );
    }
}
//---------------------------------------------------------------------------
HDC		__fastcall 	TDirectDraw::GetBackBufferDC()
{
// if the DC is already retrieved, skip out
if (fPrimarySurfaceDC != NULL) return NULL;
if (fBackBufferDC != NULL)
	{	fBackBufferDCCnt++;
    return fBackBufferDC;
    }

HDC hdc;
fErrorValue = fBackBuffer->GetDC(&hdc);
//int iter = 0;
//while ((fErrorValue == DDERR_WASSTILLDRAWING) && (++iter<1000000)) fErrorValue = fBackBuffer->GetDC(&hdc);
if (fErrorValue == DDERR_SURFACELOST)
   	{
    fPrimarySurface->Restore();
    fErrorValue = fBackBuffer->GetDC(&hdc);
    }
if (fErrorValue != DD_OK)
	{
    fBackBufferDC = NULL;
    Error(fErrorValue,"TDirectDraw::GetBackBufferDC()","Could not retrieve a DC for the back buffer");
    }
else
	{
    fBackBufferDC = hdc;
	fBackBufferDCCnt++;
//    Debug(DBG_NOTE,"TDirectDraw::GetBackBufferDC()","a DC for the back buffer was retrieved");
    }
return fBackBufferDC;
}
//---------------------------------------------------------------------------
HDC		__fastcall 	TDirectDraw::GetPrimarySurfaceDC()
{
// if the DC is already retrieved, skip out
if (fBackBufferDC != NULL) return NULL;
if (fPrimarySurfaceDC != NULL)
	{
	fPrimarySurfaceDCCnt++;
    return fPrimarySurfaceDC;
    }

HDC hdc;
fErrorValue = fPrimarySurface->GetDC(&hdc);
//int iter = 0;
//while ((fErrorValue == DDERR_WASSTILLDRAWING) && (++iter<1000000)) fErrorValue = fPrimarySurface->GetDC(&hdc);
if (fErrorValue == DDERR_SURFACELOST)
   	{
    fPrimarySurface->Restore();
    fErrorValue = fPrimarySurface->GetDC(&hdc);
    }
if (fErrorValue != DD_OK)
	{
    fPrimarySurfaceDC = NULL;
    Error(fErrorValue,"TDirectDraw::GetPrimarySurfaceDC()","Could not retrieve a DC for the primary surface");
    }
else
	{
    fPrimarySurfaceDC = hdc;
    fPrimarySurfaceDCCnt++;
//    Debug(DBG_NOTE,"TDirectDraw::GetPrimarySurfaceDC()","a DC for the primary surface was retrieved");
    }
return fPrimarySurfaceDC;
}
//---------------------------------------------------------------------------
HDC	__fastcall TDirectDraw::ReleaseBackBufferDC()
{
if (fBackBufferDC == NULL) return NULL;
if (fPrimarySurfaceDC != NULL) return NULL;

if (fBackBufferDCCnt-1 > 0)
	{
    fBackBufferDCCnt--;
    return fBackBufferDC;
    }

fErrorValue = fBackBuffer->ReleaseDC(fBackBufferDC);
if (fErrorValue != DD_OK)
	{
    Error(fErrorValue,"TDirectDraw::ReleaseBackBufferDC()","Could not release the back buffer DC"); // note: should _never_ happen
    return fBackBufferDC;
    }
fBackBufferDC = NULL;
fBackBufferDCCnt--;
return fBackBufferDC;
}
//---------------------------------------------------------------------------
HDC		__fastcall TDirectDraw::ReleasePrimarySurfaceDC()
{
if (fBackBufferDC != NULL) return NULL;
if (fPrimarySurfaceDC == NULL) return NULL;

if (fPrimarySurfaceDCCnt-1 > 0)
	{
    fPrimarySurfaceDCCnt--;
    return fPrimarySurfaceDC;
    }

fErrorValue = fPrimarySurface->ReleaseDC(fPrimarySurfaceDC);
if (fErrorValue != DD_OK)
	{
    Error(fErrorValue,"TDirectDraw::ReleasePrimarySurfaceDC()","Could not release the primary surface DC"); // note: should _never_ happen
    return fPrimarySurfaceDC;
    }
fPrimarySurfaceDC = NULL;
fPrimarySurfaceDCCnt--;
return fPrimarySurfaceDC;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::Flip(bool wait)
{
if (fDirectDraw == NULL) return false;

//
// NOTE: only surfaces with DDSCAPS_FLIP and DDSCAPS_FRONTBUFFER can be flipped
//		 ie. the primary surface
//

fErrorValue = DD_OK + 1;
if (fPrimarySurface != NULL)
   	{
    int flags = 0;
    if (wait) flags |= DDFLIP_WAIT;
    fErrorValue = fPrimarySurface->Flip(NULL,flags);
    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::Flip()","the surface could not be flipped");
//    else Debug(DBG_NOTE,"TDirectDraw::Flip()","the surface has been flipped successfully");
//    for (int adelay=0;adelay<250000000;adelay++) ;
    }
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipToGDI()
{
if (fDirectDraw == NULL) return false;

// simply call the FlipToGDISurface function & use result
fErrorValue = fDirectDraw->FlipToGDISurface();
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipToGDI()","couldn't flip to gdi surface");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipPrimarySurface(bool wait)
{
if (fDirectDraw == NULL) return false;

// simply call the BltFast function & use result
RECT r;
r.left = 0;
r.top = 0;
r.right = fWidth-1;
r.bottom = fHeight-1;
int flags = DDBLTFAST_NOCOLORKEY;
if (wait) flags |= DDBLTFAST_WAIT;
fErrorValue = fBackBuffer->BltFast(0,0,fPrimarySurface,NULL,flags);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipGDO()","couldn't flip primary to back buffer surface");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipBackBuffer(bool wait)
{
if (fDirectDraw == NULL) return false;

// simply call the BltFast function & use result
RECT r;
r.left = 0;
r.top = 0;
r.right = fWidth-1;
r.bottom = fHeight-1;

int flags = DDBLTFAST_NOCOLORKEY;
if (wait) flags |= DDBLTFAST_WAIT;
fErrorValue = fPrimarySurface->BltFast(0,0,fBackBuffer,NULL,flags);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipBack()","couldn't flip back buffer to primary surface");

#ifdef DIRECTX_MOUSE_CURSOR
UpdateMouseCursor(true);
#endif

return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipPrimarySurfaceRect(int left,int top,int right,int bottom, bool wait)
{
if (fDirectDraw == NULL) return false;

// simply call the BltFast function & use result
RECT r;
r.left = left;
r.top = top;
r.right = right;
r.bottom = bottom;

if (r.left<0) r.left = 0;
if (r.top<0) r.top = 0;
if (r.right>fWidth) r.right = fWidth;
if (r.bottom>fHeight) r.bottom = fHeight;

int flags = DDBLTFAST_NOCOLORKEY;
if (wait) flags |= DDBLTFAST_WAIT;
fErrorValue = fBackBuffer->BltFast(r.left,r.top,fPrimarySurface,&r,flags);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipPrimaryRect()","couldn't flip primary rect to back buffer");
if (fErrorValue == DDERR_INVALIDRECT) Error(fErrorValue,"TDirectDraw::FlipPrimaryRect()","rectangle requested is invalid");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipBackBufferRect(int left,int top,int right,int bottom, bool wait)
{
if (fDirectDraw == NULL) return false;

// simply call the BltFast function & use result
RECT r;
r.left = left;
r.top = top;
r.right = right;
r.bottom = bottom;

if (r.left<0) r.left = 0;
if (r.top<0) r.top = 0;
if (r.right>fWidth) r.right = fWidth;
if (r.bottom>fHeight) r.bottom = fHeight;

// and flip the area
int flags = DDBLTFAST_NOCOLORKEY;
if (wait) flags |= DDBLTFAST_WAIT;
fErrorValue = fPrimarySurface->BltFast(r.left,r.top,fBackBuffer,&r,flags);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipBackRect()","couldn't flip back buffer rect to primary surface");
if (fErrorValue == DDERR_INVALIDRECT) Error(fErrorValue,"TDirectDraw::FlipBackRect()","rectangle requested is invalid");

#ifdef DIRECTX_MOUSE_CURSOR
UpdateMouseCursor(true);
#endif

return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::PutImage(HDC hdc,int left,int top,int width,int height,TImage* image,long copymode)
{
if ((hdc == NULL) || (image == NULL)) return false;
TCanvas* canvas = new TCanvas();
if (canvas == NULL) return false;

bool result;
TRect r;
r.Top = top;
r.Left = left;
r.Right = left + width;
r.Bottom = top + height;
canvas->Handle = hdc;
try
	{
    canvas->CopyMode = copymode;
    canvas->StretchDraw(r,image->Picture->Graphic);
    result = true;
    }
catch ( ... )
	{
    result = false;
    }
canvas->Handle = NULL;
delete canvas;
return result;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::PutBackBufferImage(int left,int top,int width,int height,TImage* image,long copymode)
{
HDC hdc = GetBackBufferDC();
if (hdc != NULL)
	{
    PutImage(hdc,left,top,width,height,image,copymode);
    ReleaseBackBufferDC();
    return true;
    }
return false;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::PutPrimarySurfaceImage(int left,int top,int width,int height,TImage* image,long copymode)
{
HDC hdc = GetPrimarySurfaceDC();
if (hdc != NULL)
	{
    PutImage(hdc,left,top,width,height,image,copymode);
    ReleasePrimarySurfaceDC();
    return true;
    }
return false;
}
//---------------------------------------------------------------------------
TImage*	__fastcall	TDirectDraw::GetImage(HDC hdc,int left,int top,int width,int height, long copymode )
{
if (hdc == NULL) return NULL;

TImage* image = new TImage(this);
if (image == NULL) return NULL;

TCanvas* canvas = new TCanvas();
if (canvas == NULL) { delete image; return NULL; }

bool result;
try
	{
	canvas->Handle = hdc;

    TRect srce_r;
    srce_r.Left = left;
    srce_r.Top = top;
    srce_r.Right = left + width;
    srce_r.Bottom = top + height;

    image->AutoSize = true;
    image->BoundsRect = srce_r;
    image->Canvas->CopyMode = copymode;
    image->Canvas->CopyRect( srce_r, canvas, srce_r );

    result = true;
    }
catch ( ... )
	{
    result = false;
    }

canvas->Handle = NULL;
delete canvas;

// only return the image if everything worked smoothly
if (result) return image;

delete image;
return NULL;
}
//---------------------------------------------------------------------------
TImage*	__fastcall	TDirectDraw::GetBackBufferImage(int left,int top,int width,int height,long copymode)
{
HDC hdc = GetBackBufferDC();
if (hdc != NULL)
	{
    TImage* image = GetImage(hdc,left,top,width,height,copymode);
    ReleaseBackBufferDC();
    return image;
    }
return NULL;
}
//---------------------------------------------------------------------------
TImage*	__fastcall	TDirectDraw::GetPrimarySurfaceImage(int left,int top,int width,int height,long copymode)
{
HDC hdc = GetPrimarySurfaceDC();
if (hdc != NULL)
	{
    TImage* image = GetImage(hdc,left,top,width,height,copymode);
    ReleasePrimarySurfaceDC();
    return image;
    }
return NULL;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SetDisplayMode(int width,int height,int bitsperpixel)
{
if (fDirectDraw == NULL) return false;

// since the display mode can only be changed when in "Exclusive Mode"...
fErrorValue = DD_OK;

// make sure to release the surfaces & change the mode if and only if required
if ((width != fWidth) || (height != fHeight) || (bitsperpixel != fBitsPerPixel))
	{
    // tell DirectDraw to change the display mode to that required
	fErrorValue = fDirectDraw->SetDisplayMode(width,height,bitsperpixel);
	if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::SetDisplayMode()","could not set display mode to: " + IntToStr(width) + "x" + IntToStr(height) + "x" + IntToStr(bitsperpixel));
	else
		{
		fWidth = width;
		fHeight = height;
		fBitsPerPixel = bitsperpixel;

        // re-create the surfaces now the resolution has changed
        ReleaseSurfaces();
        CreateSurfaces();
        }
    }
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::Debug( int msgclass, System::AnsiString functiondesc, System::AnsiString msg )
{
if (!fDebugging) return;
if (fErrorCount>5000) return;
TStringList*	new_strings = new TStringList();
if (new_strings == NULL) return;	// blah

// keep track of overall memory usage
//if (overall_FirstTimeIn)
//	{
//    overall_FirstTimeIn = false;
//    InMemoryStatus( &overall_in_mstat );
//    }
//if ((msgclass == DBG_OUT) || (msgclass == DBG_IN))
//	{
//	OutMemoryStatus( &overall_out_mstat );
//	if (!SameMemoryStatus( &overall_in_mstat, &overall_out_mstat ))
//		DumpMemoryStatus( "OVERALL MEMORY USAGE", &overall_in_mstat, &overall_out_mstat );
//    }

// simple patch
System::AnsiString	indent = "";
int indenture;
for (indenture=0;indenture<fErrorIndent;indenture++) indent = indent + "    ";

switch (msgclass)
	{
    case	DBG_IN:
    		if (fErrorIndent == 0) new_strings->Add(indent + "-----------------------------------------------------------------------------");
            else new_strings->Add(indent + " - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ");
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "IN> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "IN> " + functiondesc);
            fErrorIndent ++;
            break;

    case	DBG_OUT:
			fErrorIndent --;
			indent = "";
			for (indenture=0;indenture<fErrorIndent;indenture++) indent = indent + "    ";
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "OUT> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "OUT> " + functiondesc);
    		if (fErrorIndent != 0) new_strings->Add(indent + " - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ");
            break;

    case	DBG_INFO:
            if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "INFO> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "INFO> " + functiondesc);
            break;

    case	DBG_NOTE:
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "NOTE> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "NOTE> " + functiondesc);
            break;

    case	DBG_ERROR:
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "ERROR> " + functiondesc + " --> " + msg);
    		else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "ERROR> " + functiondesc);
            break;

    }
for (int i=0;i<new_strings->Count;i++) fErrorStrings->Add(new_strings->Strings[i]);
new_strings->Clear();
delete new_strings;
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::Error( int errorvalue, System::AnsiString functiondesc, System::AnsiString errormsg )
{
Debug(DBG_ERROR,functiondesc,"Err # " + IntToStr(errorvalue) + " " + ErrorMessage(errorvalue));
Debug(DBG_ERROR,functiondesc,"Err : " + errormsg);
}
//---------------------------------------------------------------------------
System::AnsiString	__fastcall	TDirectDraw::ErrorMessage(int errorvalue)
{
System::AnsiString	errormessage = "";
switch (errorvalue)
	{
    // common directx results
    case DD_OK:						errormessage = "DD_OK, DC retrieved successfully"; break;
	case DDERR_GENERIC:             errormessage = "DDERR_GENERIC, generic error"; break;
	case DDERR_INVALIDOBJECT:		errormessage = "DDERR_INVALIDOBJECT, Invalid object"; break;
	case DDERR_INVALIDPARAMS:		errormessage = "DDERR_INVALIDPARAMS, Invalid params"; break;
	case DDERR_INVALIDSURFACETYPE:	errormessage = "DDERR_INVALIDSURFACETYPE, Invalid surface type"; break;
	case DDERR_UNSUPPORTED:			errormessage = "DDERR_UNSUPPORTED, Unsupported"; break;
	case DDERR_SURFACELOST:			errormessage = "DDERR_SURFACELOST, Surface is lost"; break;
	case DDERR_SURFACEBUSY:			errormessage = "DDERR_SURFACEBUSY, Surface is busy"; break;
	case DDERR_WASSTILLDRAWING:		errormessage = "DDERR_WASSTILLDRAWING, Was still drawing"; break;
	// flip results
	case DDERR_NOFLIPHW:			errormessage = "DDERR_NOFLIPHW, No flip hardware"; break;
	case DDERR_NOTFLIPPABLE:		errormessage = "DDERR_NOTFLIPPABLE, Surface is not flippable"; break;
    // GetDC results
	case DDERR_DCALREADYCREATED:	errormessage = "DDERR_DCALREADYCREATED, DC was already created"; break;
    default:						errormessage = "UNKNOWN ERROR, the error generated is unknown";
    };
return errormessage;
}
//---------------------------------------------------------------------------
/*void	__fastcall	TDirectDraw::Clear(TColor color)
{
Debug(DBG_IN,"TDirectDraw::Clear()","color="+IntToStr(color));
if (fBackBufferUsedLast)
	{
	Debug(DBG_INFO,"TDirectDraw::Clear()","about to clear the back buffers memory");
	memset( fBackBufferLockingDesc.lpSurface, color, fWidth * fHeight );
	Debug(DBG_INFO,"TDirectDraw::Clear()","about to flip the back buffer");
	Flip();
    }
else
    if (fDrawingOnPrimarySurface)
		{
		Debug(DBG_INFO,"TDirectDraw::Clear()","about to clear the primary surfaces memory");
		memset( fPrimarySurfaceLockingDesc.lpSurface, color, fWidth * fHeight );
		Debug(DBG_INFO,"TDirectDraw::Clear()","about to flip the primary surface");
        Flip();
	    }
Debug(DBG_OUT,"TDirectDraw::Clear()","");
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::LockSurface(LPDIRECTDRAWSURFACE surface)
{
if (fDirectDraw == NULL) return;
Debug(DBG_IN,"TDirectDraw::LockSurface()","surface=" + IntToStr((int)surface));

// setup the surface description we use for locking / unlocking surfaces
Debug(DBG_NOTE,"TDirectDraw::Constructor()","setting up the locking / unlocking surface descriptions");
memset(&fPrimarySurfaceLockingDesc,0,sizeof(fPrimarySurfaceLockingDesc));
fPrimarySurfaceLockingDesc.dwSize = sizeof(fPrimarySurfaceLockingDesc);
memset(&fBackBufferLockingDesc,0,sizeof(fBackBufferLockingDesc));
fBackBufferLockingDesc.dwSize = sizeof(fBackBufferLockingDesc);

if (surface != NULL)
	{
	// note - this function will need to be modified to deal correctly with any surface
	if ((surface != fPrimarySurface) && (surface != fBackBuffer))
    	{
        Debug(DBG_WARNING,"TDirectDraw::LockSurface()","surface cannot be locked because it is neither the primary surface or the back buffer");
        return false;
        }

    if (surface == fBackBuffer)
    	{
        if (!fBackBufferLocked)
	    	{
	        Debug(DBG_INFO,"TDirectDraw::LockSurface()","about to lock the back buffer");
	        fErrorValue = surface->Lock(NULL,&fBackBufferLockingDesc,DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,NULL);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::LockSurface()","the back buffer could not be locked");
		        return false;
	            }
            fBackBufferLocked = true;
			fBackBufferUsedLast = true;
            fPrimarySurfaceUsedLast = false;
		    Debug(DBG_NOTE,"TDirectDraw::LockSurface()","the back buffer locked successfully");
        	}
        else
		    Debug(DBG_WARNING,"TDirectDraw::LockSurface()","the back buffer is already locked");
	    }
    else
    	if (!fPrimarySurfaceLocked)
	    	{
    	    Debug(DBG_INFO,"TDirectDraw::LockSurface()","about to lock the primary surface");
	        fErrorValue = surface->Lock(NULL,&fPrimarySurfaceLockingDesc,DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,NULL);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::LockSurface()","the primary surface could not be locked");
		        return false;
	            }
            fPrimarySurfaceLocked = true;
            fBackBufferUsedLast = false;
			fPrimarySurfaceUsedLast = true;
		    Debug(DBG_NOTE,"TDirectDraw::LockSurface()","the primary surface locked successfully");
	        }
        else
		    Debug(DBG_WARNING,"TDirectDraw::LockSurface()","the primary surface is already locked");
	return true;
    }
Debug(DBG_OUT,"TDirectDraw::LockSurface()","");
return false;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::UnLockSurface(LPDIRECTDRAWSURFACE surface)
{
if (fDirectDraw == NULL) return;
Debug(DBG_IN,"TDirectDraw::UnLockSurface()","surface=" + IntToStr((int)surface));
if (surface != NULL)
	{
	// note - this function will need to be modified to deal correctly with any surface
	if ((surface != fPrimarySurface) && (surface != fBackBuffer))
    	{
        Debug(DBG_WARNING,"TDirectDraw::UnLockSurface()","surface cannot be unlocked because it is neither the primary surface or the back buffer");
        return false;
        }

    if (surface == fBackBuffer)
    	{
        if (fBackBufferLocked)
        	{
	        Debug(DBG_INFO,"TDirectDraw::UnLockSurface()","about to lock the back buffer");
			surface->Unlock(fBackBufferLockingDesc.lpSurface);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::UnLockSurface()","the back buffer could not be unlocked");
		        return false;
	            }
            fBackBufferLocked = false;
		    Debug(DBG_NOTE,"TDirectDraw::UnLockSurface()","the back buffer unlocked successfully");
            }
        else
		    Debug(DBG_WARNING,"TDirectDraw::UnLockSurface()","the back buffer was already unlocked");
	    }
    else
    	if (fPrimarySurfaceLocked)
	    	{
	        Debug(DBG_INFO,"TDirectDraw::UnLockSurface()","about to lock the primary surface");
			fErrorValue = surface->Unlock(fPrimarySurfaceLockingDesc.lpSurface);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::UnLockSurface()","the primary surface could not be unlocked");
		        return false;
	            }
	        fPrimarySurfaceLocked = false;
		    Debug(DBG_NOTE,"TDirectDraw::UnLockSurface()","the primary surface unlocked successfully");
	        }
        else
		    Debug(DBG_NOTE,"TDirectDraw::UnLockSurface()","the primary surface was already unlocked");
	return true;
    }
Debug(DBG_OUT,"TDirectDraw::UnLockSurface()","");
return false;
}
*/
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::SetMouseCursor( HICON cursor )
{
if (fDirectDraw == NULL) return;
if (fMouseBusy!=NULL) return;
fMouseCursor = cursor;
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::SetMousePos( int x,int y )
{
if (fDirectDraw == NULL) return;
if (fMouseBusy!=NULL) return;

fMousePos.x = x;
fMousePos.y = y;
UpdateMouseCursor(true);
}
//---------------------------------------------------------------------------
void __fastcall	TDirectDraw::UpdateMouseCursor(bool forced)
{
#ifdef DIRECTX_MOUSE_CURSOR
if (fDirectDraw == NULL) return;
if (fMouseBusy!=NULL) return;

// determine the current coordinates of the mouse cursor
POINT p;
GetCursorPos( &p );

// if the mouse has moved
if ((p.x != fMousePos.x) || (p.y != fMousePos.y) || forced)
	{
    // remember the new coordinates
   	fMousePos = p;

	// determine the screen rectangle underneath the soon to be painted mouse cursor icon
	fMouseNewRect.Top = fMousePos.y;
	fMouseNewRect.Left = fMousePos.x;
	fMouseNewRect.Right = fMouseNewRect.Left + fMouseWidth;
	fMouseNewRect.Bottom = fMouseNewRect.Top + fMouseHeight;

	// determine the size of the total bitmap, which is defined by the minimum size
	// required to draw both the area underneath the current mouse cursor and the
	// area underneath the previous mouse cursor. all coords are screen coords
	TRect bitmap_r = Rect( min(fMouseNewRect.Left, fMouseOldRect.Left),
						min(fMouseNewRect.Top, fMouseOldRect.Top),
                        max(fMouseNewRect.Right, fMouseOldRect.Right),
                        max(fMouseNewRect.Bottom, fMouseOldRect.Bottom) );

	// only paint the cursor if a cursor exists
	if (fMouseCursor != NULL)
		{
		// setup for drawing onto the primary surface (the visible display)
		TCanvas* canvas = BeginPrimarySurfaceDrawing( bitmap_r.Left, bitmap_r.Top, bitmap_r.Right, bitmap_r.Bottom );

		// iff the surface is okay
		if (canvas != NULL)
		   	{
	        // the very first time we attempt to draw the mouse, create the underneath surface
	        if (fMouseFirstTimeDrawn)
	        	{
				fMouseUnder = CreateCompatibleBitmap( canvas->Handle, fMouseWidth, fMouseHeight );
	            fMouseUnderCursor->Handle = fMouseUnder;
	            }

	        // create a couple of temporary vars
	        int bitmap_width = bitmap_r.Right-bitmap_r.Left;
	        int bitmap_height = bitmap_r.Bottom-bitmap_r.Top;

	        // setup the temporary drawing surface so that we blit once and avoid flicker
			HBITMAP hbitmap = CreateCompatibleBitmap( canvas->Handle, bitmap_width, bitmap_height );
	        fMouseBitmap->Handle = hbitmap;

	        // determine the internal rectangles used by both the new & old mouse cursors
	        TRect local_mouse_r = Rect( 0,0, fMouseWidth, fMouseHeight );
	        TRect local_bitmap_r = Rect( 0,0, bitmap_width, bitmap_height );

	        // figure out where inside the fMouseBitmap we should be putting the old & new cursors
	        TRect old_r;
	        TRect new_r;
	        if (fMouseNewRect.Left < fMouseOldRect.Left) {
	            if (fMouseNewRect.Top > fMouseOldRect.Top) { old_r.Left = bitmap_width-fMouseWidth; old_r.Top = 0; new_r.Left = 0; new_r.Top = bitmap_height-fMouseHeight; }
	            if (fMouseNewRect.Top == fMouseOldRect.Top) { old_r.Left = bitmap_width-fMouseWidth; old_r.Top = 0; new_r.Left = 0; new_r.Top = 0; }
	            if (fMouseNewRect.Top < fMouseOldRect.Top) { old_r.Left = bitmap_width-fMouseWidth; old_r.Top = bitmap_height-fMouseHeight; new_r.Left = 0; new_r.Top = 0; }
	            }
			if (fMouseNewRect.Left == fMouseOldRect.Left) {
	            if (fMouseNewRect.Top > fMouseOldRect.Top) { old_r.Left = 0; old_r.Top = 0; new_r.Left = 0; new_r.Top = bitmap_height-fMouseHeight; }
    	        if (fMouseNewRect.Top == fMouseOldRect.Top) { old_r.Left = 0; old_r.Top = 0; new_r.Left = 0; new_r.Top = 0; }
	            if (fMouseNewRect.Top < fMouseOldRect.Top) { old_r.Left = 0; old_r.Top = bitmap_height-fMouseHeight; new_r.Left = 0; new_r.Top = 0; }
		        }
		    if (fMouseNewRect.Left > fMouseOldRect.Left) {
	            if (fMouseNewRect.Top > fMouseOldRect.Top) { old_r.Left = 0; old_r.Top = 0; new_r.Left = bitmap_width-fMouseWidth; new_r.Top = bitmap_height-fMouseHeight; }
	            if (fMouseNewRect.Top == fMouseOldRect.Top) { old_r.Left = 0; old_r.Top = 0; new_r.Left = bitmap_width-fMouseWidth; new_r.Top = 0; }
	            if (fMouseNewRect.Top < fMouseOldRect.Top) { old_r.Left = 0; old_r.Top = bitmap_height-fMouseHeight; new_r.Left = bitmap_width-fMouseWidth; new_r.Top = 0; }
	            }
			old_r.Right = old_r.Left + fMouseWidth;
			old_r.Bottom = old_r.Top + fMouseHeight;
	        new_r.Right = new_r.Left + fMouseWidth;
	        new_r.Bottom = new_r.Top + fMouseHeight;

			// --- copy exactly whats on the screen (under both old & new cursors) to the fMouseBitmap
			fMouseBitmap->Canvas->CopyRect( local_bitmap_r, canvas, bitmap_r );

			// --- except for the very first time the cursor is drawn,
			if (fMouseFirstTimeDrawn) fMouseFirstTimeDrawn = false;
			else
		        // erase the existing mouse cursor from within the offscreen fMouseBitmap
				fMouseBitmap->Canvas->CopyRect( old_r, fMouseUnderCursor->Canvas, local_mouse_r );

	        // --- save what is underneath the new mouse cursor location (taken from fMouseBitmap)
            if ((new_r.Left != old_r.Left) || (new_r.Top != old_r.Top))
		       	fMouseUnderCursor->Canvas->CopyRect( local_mouse_r, fMouseBitmap->Canvas, new_r );

		    // --- paint the cursor at the new position within fMouseBitmap
//	        fMouseBitmap->Canvas->Pen->Color = clRed;
//	       	fMouseBitmap->Canvas->Pen->Style = psSolid;
//	       	fMouseBitmap->Canvas->Brush->Style = bsClear;
//	       	fMouseBitmap->Canvas->Rectangle( new_r.Left, new_r.Top, new_r.Right, new_r.Bottom );
			DrawIconEx( fMouseBitmap->Canvas->Handle, new_r.Left, new_r.Top, fMouseCursor, 0,0,0,NULL,DI_NORMAL);

	        // --- draw the finished bitmap on the screen
	        canvas->Draw( bitmap_r.Left, bitmap_r.Top, fMouseBitmap );

	        // --- clean up
	        fMouseBitmap->Handle = NULL;
	        DeleteObject( hbitmap );
		    }
		// clean up
		EndPrimarySurfaceDrawing( canvas );
		}
	// --- remember where we just painted
	fMouseOldRect = fMouseNewRect;
    }
//
DebugMouseCursor();
#endif
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::DebugMouseCursor()
{
TCanvas* canvas = BeginPrimarySurfaceDrawing(0,0,100,200);
if (canvas)
	{
	TRect r;

    r = Rect(0,0,100,20);
    canvas->Font->Color = clRed;
    canvas->Brush->Color = clBlack;
    canvas->TextRect(r, r.Left, r.Top, "Underneath the Cursor" );
    r = Rect(0,20,100,120);
    canvas->StretchDraw(r, fMouseUnderCursor );
    }
EndPrimarySurfaceDrawing( canvas );
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::PaintMouseCursor()
{
#ifdef DIRECTX_MOUSE_CURSOR
if (fDirectDraw == NULL) return;
if (fMouseBusy!=NULL) return;

// get the current coordinates of the mouse cursor
GetCursorPos( &fMousePos );

// determine the screen rectangle underneath the soon to be painted mouse cursor icon
fMouseNewRect.Top = fMousePos.y;
fMouseNewRect.Left = fMousePos.x;
fMouseNewRect.Right = fMouseNewRect.Left + fMouseWidth;
fMouseNewRect.Bottom = fMouseNewRect.Top + fMouseHeight;

// only paint the cursor if a cursor exists
if (fMouseCursor != NULL)
	{
	// setup for drawing onto the primary surface
	TCanvas* canvas = BeginPrimarySurfaceDrawing( fMousePos.x, fMousePos.y, fMousePos.x+fMouseWidth, fMousePos.y+fMouseHeight );

	// iff the surface is okay
	if (canvas != NULL)
	   	{
        // save what's underneath the cursor right now
        TRect screen_r = Rect( fMousePos.x, fMousePos.y, fMousePos.x+fMouseWidth, fMousePos.y+fMouseHeight );
		fMouseUnderCursor->Canvas->CopyRect( fMouseUnderCursor->Canvas->ClipRect, canvas, screen_r );
//        int x,y,xx,yy;
//       	for (x=0,xx=fMousePos.x;x<fMouseWidth;x++,xx++)
//	        for (y=0,yy=fMousePos.y;y<fMouseWidth;y++,yy++)
//            	fMouseUnderCursor->Canvas->Pixels[x][y] = canvas->Pixels[xx][yy];

        // and draw the icon over the top
//        canvas->Pen->Color = clGreen;
//        canvas->Pen->Style = psSolid;
//        canvas->Brush->Style = bsClear;
//        canvas->Rectangle( fMousePos.x, fMousePos.y, fMousePos.x+fMouseWidth, fMousePos.y+fMouseHeight );
		DrawIconEx( canvas->Handle, fMousePos.x, fMousePos.y, fMouseCursor, 0,0,0,NULL,DI_NORMAL);
        }
	// clean up
	EndPrimarySurfaceDrawing( canvas );
    }
//
DebugMouseCursor();
#endif
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::UnpaintMouseCursor()
{
#ifdef DIRECTX_MOUSE_CURSOR
if (fDirectDraw == NULL) return;
if (fMouseBusy!=NULL) return;

// only paint the cursor if a cursor exists
if (fMouseCursor != NULL)
	{
	// setup for drawing onto the primary surface
	TCanvas* canvas = BeginPrimarySurfaceDrawing( fMousePos.x, fMousePos.y, fMousePos.x+fMouseWidth, fMousePos.y+fMouseHeight );

	// iff the surface is okay
	if (canvas != NULL)
	   	{
        // restore what's underneath the cursor right now
        TRect screen_r = Rect( fMousePos.x, fMousePos.y, fMousePos.x+fMouseWidth, fMousePos.y+fMouseHeight );
        canvas->CopyRect( screen_r, fMouseUnderCursor->Canvas, fMouseUnderCursor->Canvas->ClipRect );
//        int x,y,xx,yy;
//       	for (x=0,xx=fMousePos.x;x<fMouseWidth;x++,xx++)
//	        for (y=0,yy=fMousePos.y;y<fMouseWidth;y++,yy++)
//            	canvas->Pixels[xx][yy] = fMouseUnderCursor->Canvas->Pixels[x][y];
        }
	// clean up
	EndPrimarySurfaceDrawing( canvas );
    }
//
DebugMouseCursor();
#endif
}
//---------------------------------------------------------------------------


