//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <vcl\registry.hpp>
#include "winresrc.h"
#include "penwin.h"
#include "TDirectDrawForm.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
TDirectDrawForm *DirectDrawForm;
//---------------------------------------------------------------------------
// note: the following macros need to be inserted before release so that
//       our private methods/vars do not even hint towards what we are doing.
//
// eg. 
// #define		fBorderX				var1
// #define		fBorderY				var2		// and so on
// #define		PaintFrame				func1
// #define		PaintCloseButton		func2		// and so on
//
//	- the end result being that anyone using TDirectDrawForm.H can only
//	see var1, var2, func1, func2, etc. and internally we get to use
//	fBorderX, fBorderY, PaintFrame, PaintCloseButton, etc.
//
//---------------------------------------------------------------------------
// note:	The paint code used to draw everything uses physical coordinates
//			ie. left+8, top+18, etc. for painting.
//
//			The painting should really be done using formulas, but it was quicker
//			initially to do it this way rather than spending the time figuring
//			out the correct formulas / constants.
//
//			Similarly, lots of the code is similar - it's heaps quicker and much
//			simpler to cut & paste & make huge code rather than spending weeks
//			making tight & hard to understand code.
//
//			fix any possible problems later.
//---------------------------------------------------------------------------
// note:	at some point, figure out if setting realtime priority has any effect.
// SetPriorityClass(hProcess: THandle; dwPriorityClass: DWORD): BOOL; stdcall;
//---------------------------------------------------------------------------
__fastcall TDirectDrawForm::TDirectDrawForm(TComponent* Owner)
	: TForm(Owner)
{
// determine if this pc is :-
//	(1) a standard windows 95 or NT pc (default),
//	(2) a windows 95 pc width Microsoft Plus! installed

TRegistry*	registry = new TRegistry();
if (registry!=NULL)
	{
    // make sure the root is set to HKEY_LOCAL_MACHINE
    HKEY orig_rootkey = registry->RootKey;
    registry->RootKey = HKEY_LOCAL_MACHINE;

    // check if microsoft plus! is installed
    registry->OpenKey("SOFTWARE\\Microsoft\\Windows\\CurrentVersion", false );
    System::AnsiString	PlusVersionNumber = registry->ReadString("Plus! VersionNumber");
    registry->CloseKey();
    fMicrosoftPlus = (PlusVersionNumber != "");
//		if (fMicrosoftPlus) ShowMessage("This is a Windows 95 pc running Microsoft Plus!");
//		else ShowMessage("This is a standard Windows 95 or NT pc");

	// clean up
    registry->RootKey = orig_rootkey;
    delete registry;
    }

// get the width, height of the minimized bar
MINIMIZEDMETRICS minmetrics;
minmetrics.cbSize = sizeof(MINIMIZEDMETRICS);
SystemParametersInfo( SPI_GETMINIMIZEDMETRICS, sizeof(MINIMIZEDMETRICS), &minmetrics, 0 );

fMinimizedWidth = minmetrics.iWidth;
fBorderX = GetSystemMetrics(SM_CXDLGFRAME);
fBorderY = GetSystemMetrics(SM_CYDLGFRAME);
fCaptionHeight = GetSystemMetrics(SM_CYCAPTION);
fButtonWidth = GetSystemMetrics(SM_CXSMICON);
fButtonHeight = GetSystemMetrics(SM_CYSMICON);
fScrollbarWidth = GetSystemMetrics( SM_CXHSCROLL );
fScrollbarHeight = GetSystemMetrics( SM_CYHSCROLL );
fThumbnailHeight = GetSystemMetrics( SM_CXHTHUMB );
fIconWidth = GetSystemMetrics( SM_CXSMICON );
fIconHeight = GetSystemMetrics( SM_CYSMICON );

fFormIsActive = false;
fFormIsMoving = false;
fFormIsMoveable = true;
fFormIsBeingRestored = false;
fFormIsBeingMinimized = false;
fFormIsBeingMaximized = false;
fMainFormIsMinimized = false;
fMenuIsDisplayed = false;

fHandlingWMPAINT = false;
fFramesFlipped = 0;
fTransparent = false;

fHScrollBarIsVisible = true;
fVScrollBarIsVisible = true;

// save the name of the font used to render the caption
NONCLIENTMETRICS ncm;
ncm.cbSize = sizeof(NONCLIENTMETRICS);
SystemParametersInfo( SPI_GETNONCLIENTMETRICS, 0, &ncm, 0 );
fCaptionFontName = ncm.lfCaptionFont.lfFaceName;
fSmCaptionFontName = ncm.lfSmCaptionFont.lfFaceName;

// pre-load all required bitmaps
fUpArrow = new Graphics::TBitmap();
fDnArrow = new Graphics::TBitmap();
fLtArrow = new Graphics::TBitmap();
fRtArrow = new Graphics::TBitmap();
if (fUpArrow) { fUpArrow->Handle = LoadBitmap( NULL, MAKEINTRESOURCE(OBM_UPARROW) ); fUpArrow->IgnorePalette = true; }
if (fDnArrow) { fDnArrow->Handle = LoadBitmap( NULL, MAKEINTRESOURCE(OBM_DNARROW) ); fDnArrow->IgnorePalette = true; }
if (fLtArrow) { fLtArrow->Handle = LoadBitmap( NULL, MAKEINTRESOURCE(OBM_LFARROW) ); fLtArrow->IgnorePalette = true; }
if (fRtArrow) { fRtArrow->Handle = LoadBitmap( NULL, MAKEINTRESOURCE(OBM_RGARROW) ); fRtArrow->IgnorePalette = true; }

// pre-render the application's small icon (it doesn't seem to be supplied by the VCL at all!)
// 	ie. convert the TIcon to a TImageList so we can draw it transparently.

// lookit all of the code! --- really should create a TGraphicsFX object with methods like
//	TGraphicsFX::IconToImageList()
//	TGraphicsFX::ImageToIcon()
// or similar.
fSmallIcon = new TImageList(this);
if (fSmallIcon)
	{
    HICON large = NULL;
    HICON small = NULL;
	int result = ExtractIconEx( Application->ExeName.c_str(), 0, &large, &small, 1 );
	if (result>1)
    	{
        Graphics::TIcon* ticon = new Graphics::TIcon();
        if (ticon)
        	{
            Graphics::TBitmap* normal_tbitmap = new Graphics::TBitmap();
            if (normal_tbitmap)
            	{
                Graphics::TBitmap* small_tbitmap = new Graphics::TBitmap();
                if (small_tbitmap)
                	{
                    // use the icon retrieved from the file
		            ticon->Handle = small;

                    // copy it to a bitmap (so that it can be stretched)
	                normal_tbitmap->IgnorePalette = true;
	                normal_tbitmap->Width = ticon->Width;
	                normal_tbitmap->Height = ticon->Height;
	                normal_tbitmap->Canvas->Draw( 0, 0, ticon );

                    // shrink the bitmap down to it's required size
                    small_tbitmap->IgnorePalette = true;
                    small_tbitmap->Width = fIconWidth;
                    small_tbitmap->Height = fIconHeight;
                    TRect small_r = Rect( 0, 0, fIconWidth, fIconHeight );
                    small_tbitmap->Canvas->StretchDraw( small_r, normal_tbitmap );

                    // and turn it into a TBitmapList so we can draw it transparently
		            fSmallIcon->DrawingStyle = dsTransparent;
			        fSmallIcon->Width = small_tbitmap->Width;
			        fSmallIcon->Height = small_tbitmap->Height;
			        fSmallIcon->AddMasked( small_tbitmap, clWhite );

                    delete small_tbitmap;
	                delete normal_tbitmap;
		            delete ticon;
                	}
                else
                	{
                    delete normal_tbitmap;
                    delete ticon;
                    }
                }
            else
            	{
                delete ticon;
                }
            }
        }
	DeleteObject( small );
    DeleteObject( large );
    }

// if DirectDraw doesn't exist already, try creating it
if (DirectDraw == NULL)
	{
    DirectDraw = new TDirectDraw(this);
    // if DirectDraw was successfully created by this form...
    if (DirectDraw != NULL)
    	{
        // attempt to overload the Application.OnIdle() method with our own mouse code
        Application->OnIdle = OnIdle;

        // capture all mouse events and send them to this window
//        HWND orig_window = SetCapture( Handle );

		DirectDraw->Debugging = true;
        }
	else
    	{
    	ShowMessage( "This application could not run because TDirectDraw could not be created." );
        Application->Terminate();
        }
    }
if (DirectDraw != NULL)
	{
	fDisplayWidth = DirectDraw->Width;
    fDisplayHeight = DirectDraw->Height;
	fDisplayBitsPerPixel = DirectDraw->BitsPerPixel;
	}
}
//---------------------------------------------------------------------------
__fastcall 	TDirectDrawForm::~TDirectDrawForm()
{
if (fSmallIcon) delete fSmallIcon;
if (fRtArrow) delete fRtArrow;
if (fLtArrow) delete fLtArrow;
if (fDnArrow) delete fDnArrow;
if (fUpArrow) delete fUpArrow;

if (DirectDraw != NULL)
	if (DirectDraw->Owner == this)
		{
//        ReleaseCapture();
	    delete DirectDraw;
	    DirectDraw = NULL;
	    }
}
//---------------------------------------------------------------------------
void	__fastcall		TDirectDrawForm::HandleException(TObject *Sender)
{
if (DirectDraw != NULL)
	{
    DirectDraw->FlipToGDI();
    delete DirectDraw;
    DirectDraw = NULL;
    }
ShowMessage( "Something pretty serious caused an exception. Application has terminated abnormally." );
Application->Terminate();
}
//---------------------------------------------------------------------------
void	__fastcall		TDirectDrawForm::OnIdle(TObject* Sender, bool &done)
{
if (DirectDraw!=NULL) DirectDraw->UpdateMouseCursor();
}
//---------------------------------------------------------------------------
RECT	__fastcall	TDirectDrawForm::SelectEntireForm( HDC hdc )
{
// determine the bounds of the frame that we have to paint
RECT form_r;

// when the window is minimized, retrieve relevant sizes,
// and overwrite our internal location & size variables that are
// setup each time we get a WM_SIZE or WM_MOVE message.
if (WindowState == wsMinimized)
	{
	// get the top, left of the minimized bar
	WINDOWPLACEMENT wndplace;
	wndplace.length = sizeof(WINDOWPLACEMENT);
	GetWindowPlacement( Handle, &wndplace );

	form_r.left = wndplace.ptMinPosition.x;
	form_r.top = wndplace.ptMinPosition.y;
	form_r.right = form_r.left + fMinimizedWidth + 2*fBorderX;
	form_r.bottom = form_r.top + fBorderY + fCaptionHeight + 2;
	}
else
	{
	form_r.left = Left;
	form_r.top  = Top;
	form_r.right = form_r.left+Width;
	form_r.bottom = form_r.top+Height;
    }

// tell windows that the region of the whole form (ie. including frame)
// is to be repainted
InvalidateRect(Handle,&form_r,false);

// set the clip rect so that all drawing is clipped to the form area
HRGN form_hrgn = CreateRectRgn( form_r.left, form_r.top, form_r.right, form_r.bottom );
SelectClipRgn( hdc, form_hrgn );
DeleteObject( form_hrgn );

// make sure not to paint over any other forms, etc.
ExcludeOtherForms( Handle, hdc );

// return the resulting form rect
return form_r;
}
//---------------------------------------------------------------------------
RECT	__fastcall	TDirectDrawForm::SelectClientArea( HDC hdc, RECT form_r )
{
RECT client_r = form_r;
// only paint internals when the form is not minimized
if (WindowState != wsMinimized)
	{
	// determine the client area just inside the frame
	switch (BorderStyle)
    	{
        case bsNone:	break;
        case bsDialog:
            if (fMicrosoftPlus) client_r.top += 24;
            else client_r.top += 22;
            client_r.left += 3;
            client_r.right -= 3;
            client_r.bottom -= 3;
            break;
        case bsSingle:
            if (fMicrosoftPlus) client_r.top += 24;
            else client_r.top += 22;
            client_r.left += 3;
            client_r.right -= 3;
            client_r.bottom -= 3;
            break;
        case bsSizeable:
            if (fMicrosoftPlus) client_r.top += 25;
            else client_r.top += 23;
            client_r.left += 4;
            client_r.right -= 4;
            client_r.bottom -= 4;
            break;
        case bsSizeToolWin:
            if (fMicrosoftPlus) client_r.top += 23;
            else client_r.top += 20;
            client_r.left += 4;
            client_r.right -= 4;
            client_r.bottom -= 4;
            break;
        case bsToolWindow:
            if (fMicrosoftPlus) client_r.top += 22;
            else client_r.top += 19;
        	client_r.left += 3;
            client_r.right -= 3;
            client_r.bottom -= 3;
        	break;
        }

	// exclude other elements of frame
    if ((BorderStyle==bsSizeable) || (BorderStyle==bsSizeToolWin))
    	{
		if (fVScrollBarIsVisible) client_r.right -= fScrollbarWidth;
    	if (fHScrollBarIsVisible) client_r.bottom -= fScrollbarHeight;
        }

	// adjust the clip rect so that all drawing is clipped to the client area
	HRGN client_hrgn = CreateRectRgn( client_r.left, client_r.top, client_r.right, client_r.bottom );
	SelectClipRgn( hdc, client_hrgn );
	DeleteObject( client_hrgn );

	// make sure not to paint over any other forms
    ExcludeOtherForms( Handle, hdc );

    // make sure that we don't paint over any menus
//	ExcludeMenu( hdc, NULL );
	}
return client_r;
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintWindowFrame( HDC hdc, RECT form_r )
{
// check for prerequisites & skip out if invalid
if (hdc == NULL) return;
if (DirectDraw == NULL) return;

// set the canvas to paint on the supplied DC
Canvas->Handle = hdc;
try
	{
    switch (BorderStyle)
    	{
        case bsNone:		break;
        case bsDialog:		PaintDialogFrame( form_r ); break;
        case bsSingle:		PaintSingleFrame( form_r ); break;
        case bsSizeable:	PaintSizeableFrame( form_r ); break;
        case bsSizeToolWin:	PaintSizeToolWinFrame( form_r ); break;
        case bsToolWindow:	PaintToolWindowFrame( form_r ); break;
        }
    //PaintSystemMenu();
    }
catch ( ... ) {}
Canvas->Handle = NULL;
}
//---------------------------------------------------------------------------
void 	__fastcall	TDirectDrawForm::PaintDialogFrame( RECT paint_r )
{
// modify supplied coords for pixel rounding errors
paint_r.right--;
paint_r.bottom--;

// draw basic frame
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnFace;
for (int i=0;i<3;i++)
	{
	Canvas->MoveTo( paint_r.left+i, paint_r.top+i );
	Canvas->LineTo( paint_r.right-i, paint_r.top+i );
	Canvas->LineTo( paint_r.right-i, paint_r.bottom-i );
	Canvas->LineTo( paint_r.left+i, paint_r.bottom-i );
	Canvas->LineTo( paint_r.left+i, paint_r.top+i );
    }
// highlight left & top
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnHighlight;
Canvas->MoveTo( paint_r.left+1, paint_r.bottom );
Canvas->LineTo( paint_r.left+1, paint_r.top+1 );
Canvas->LineTo( paint_r.right, paint_r.top+1 );
// darken right & bottom
Canvas->Pen->Color = cl3DDkShadow;
Canvas->MoveTo( paint_r.right, paint_r.top );
Canvas->LineTo( paint_r.right, paint_r.bottom );
Canvas->LineTo( paint_r.left, paint_r.bottom );
// add extra highlight to right & bottom
Canvas->Pen->Color = clBtnShadow;
Canvas->MoveTo( paint_r.right-1, paint_r.top+1 );
Canvas->LineTo( paint_r.right-1, paint_r.bottom-1 );
Canvas->LineTo( paint_r.left+1, paint_r.bottom-1 );

// fill the background caption
Canvas->Pen->Style = psClear;
Canvas->Pen->Color = clBlack;
Canvas->Brush->Style = bsSolid;
if (fFormIsActive) Canvas->Brush->Color = clActiveCaption;
else Canvas->Brush->Color = clInactiveCaption;
if (fMicrosoftPlus) Canvas->Rectangle( paint_r.left+3, paint_r.top+3, paint_r.right-1, paint_r.top+24 );
else Canvas->Rectangle( paint_r.left+3, paint_r.top+3, paint_r.right-1, paint_r.top+22 );

// draw line between caption bar and main draw area
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnFace;
if (fMicrosoftPlus)
	{
	Canvas->MoveTo( paint_r.left+3, paint_r.top+23 );
	Canvas->LineTo( paint_r.right-2, paint_r.top+23 );
   }
else
	{
	Canvas->MoveTo( paint_r.left+3, paint_r.top+21 );
	Canvas->LineTo( paint_r.right-2, paint_r.top+21 );
   }

// move things a bit
paint_r.top += 5;
paint_r.right -= 4;

// if there is a systemmenu icon available
if (BorderIcons.Contains(biSystemMenu))
	{
	// paint the close window ("X") button
    paint_r.right -= fIconWidth;
    PaintCloseButton( paint_r.right, paint_r.top );
    paint_r.right -= 2;
    }

// draw the caption text
paint_r.left += 5;
if (fMicrosoftPlus)
	{
   paint_r.top --;
	paint_r.bottom = paint_r.top + 18;
   }
else
   paint_r.bottom = paint_r.top + 14;

Canvas->Brush->Style = bsClear;
Canvas->Font->Height = paint_r.bottom-paint_r.top;
Canvas->Font->Style = Canvas->Font->Style << fsBold;
Canvas->Font->Name = fCaptionFontName;
if (fFormIsActive) Canvas->Font->Color = clCaptionText;
else Canvas->Font->Color = clInactiveCaptionText;

// note: need to abbreviate caption so that if too large it truncates string, appends "..."
TRect tpaint_r = Rect( paint_r.left, paint_r.top, paint_r.right, paint_r.bottom );
Canvas->TextRect( tpaint_r, tpaint_r.Left, tpaint_r.Top, Caption );
}
//---------------------------------------------------------------------------
void 	__fastcall	TDirectDrawForm::PaintSingleFrame( RECT paint_r )
{
// modify supplied coords for pixel rounding errors
paint_r.right--;
paint_r.bottom--;

// draw basic frame
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnFace;
for (int i=0;i<3;i++)
	{
	Canvas->MoveTo( paint_r.left+i, paint_r.top+i );
	Canvas->LineTo( paint_r.right-i, paint_r.top+i );
	Canvas->LineTo( paint_r.right-i, paint_r.bottom-i );
	Canvas->LineTo( paint_r.left+i, paint_r.bottom-i );
	Canvas->LineTo( paint_r.left+i, paint_r.top+i );
    }
// highlight left & top
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnHighlight;
Canvas->MoveTo( paint_r.left+1, paint_r.bottom );
Canvas->LineTo( paint_r.left+1, paint_r.top+1 );
Canvas->LineTo( paint_r.right, paint_r.top+1 );
// darken right & bottom
Canvas->Pen->Color = cl3DDkShadow;
Canvas->MoveTo( paint_r.right, paint_r.top );
Canvas->LineTo( paint_r.right, paint_r.bottom );
Canvas->LineTo( paint_r.left, paint_r.bottom );
// add extra highlight to right & bottom
Canvas->Pen->Color = clBtnShadow;
Canvas->MoveTo( paint_r.right-1, paint_r.top+1 );
Canvas->LineTo( paint_r.right-1, paint_r.bottom-1 );
Canvas->LineTo( paint_r.left+1, paint_r.bottom-1 );

// fill the background caption
Canvas->Pen->Style = psClear;
Canvas->Pen->Color = clBlack;
Canvas->Brush->Style = bsSolid;
if (fFormIsActive) Canvas->Brush->Color = clActiveCaption;
else Canvas->Brush->Color = clInactiveCaption;
if (fMicrosoftPlus) Canvas->Rectangle( paint_r.left+3, paint_r.top+3, paint_r.right-1, paint_r.top+24 );
else Canvas->Rectangle( paint_r.left+3, paint_r.top+3, paint_r.right-1, paint_r.top+22 );

// draw line between caption bar and main draw area
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnFace;
if (fMicrosoftPlus)
	{
	Canvas->MoveTo( paint_r.left+3, paint_r.top+23 );
	Canvas->LineTo( paint_r.right-2, paint_r.top+23 );
   }
else
	{
	Canvas->MoveTo( paint_r.left+3, paint_r.top+21 );
	Canvas->LineTo( paint_r.right-2, paint_r.top+21 );
   }

// move things a bit
paint_r.top += 5;
paint_r.right -= 4;

// if there is a systemmenu icon available
if (BorderIcons.Contains(biSystemMenu))
	{
	// paint the close window ("X") button
    paint_r.right -= fIconWidth;
    PaintCloseButton( paint_r.right, paint_r.top );
    paint_r.right -= 2;

    // if either the maximize or minimize icons are available, _must_ draw both
    if (BorderIcons.Contains(biMaximize) || BorderIcons.Contains(biMinimize))
		{
		// paint the maximize window ("_") button
	    paint_r.right -= fIconWidth;
		if (WindowState==wsMaximized) PaintRestoreButton( paint_r.right, paint_r.top );
		else PaintMaximizeButton( paint_r.right, paint_r.top );

		// paint the minimize window ("_") button
		paint_r.right -= fIconWidth;
		if (WindowState==wsMinimized) PaintRestoreButton( paint_r.right, paint_r.top );
		else PaintMinimizeButton( paint_r.right, paint_r.top );
	    }

	// draw the "command" button
	PaintCommandButton( paint_r.left+5, paint_r.top-1 );
	paint_r.left += fIconWidth;
    }

// draw the caption text
paint_r.left += 7;
if (fMicrosoftPlus)
	{
   paint_r.top --;
	paint_r.bottom = paint_r.top + 18;
   }
else
	paint_r.bottom = paint_r.top + 14;

Canvas->Brush->Style = bsClear;
Canvas->Font->Height = paint_r.bottom-paint_r.top;
Canvas->Font->Style = Canvas->Font->Style << fsBold;
Canvas->Font->Name = fCaptionFontName;
if (fFormIsActive) Canvas->Font->Color = clCaptionText;
else Canvas->Font->Color = clInactiveCaptionText;

// note: need to abbreviate caption so that if too large it truncates string, appends "..."
TRect tpaint_r = Rect( paint_r.left, paint_r.top, paint_r.right, paint_r.bottom );
Canvas->TextRect( tpaint_r, tpaint_r.Left, tpaint_r.Top, Caption );
}
//---------------------------------------------------------------------------
void 	__fastcall	TDirectDrawForm::PaintSizeableFrame( RECT paint_r )
{
// modify supplied coords for pixel rounding errors
paint_r.right--;
paint_r.bottom--;

// draw basic frame
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnFace;
for (int i=0;i<4;i++)
	{
	Canvas->MoveTo( paint_r.left+i, paint_r.top+i );
	Canvas->LineTo( paint_r.right-i, paint_r.top+i );
	Canvas->LineTo( paint_r.right-i, paint_r.bottom-i );
	Canvas->LineTo( paint_r.left+i, paint_r.bottom-i );
	Canvas->LineTo( paint_r.left+i, paint_r.top+i );
    }
// highlight left & top
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnHighlight;
Canvas->MoveTo( paint_r.left+1, paint_r.bottom );
Canvas->LineTo( paint_r.left+1, paint_r.top+1 );
Canvas->LineTo( paint_r.right, paint_r.top+1 );
// darken right & bottom
Canvas->Pen->Color = cl3DDkShadow;
Canvas->MoveTo( paint_r.right, paint_r.top );
Canvas->LineTo( paint_r.right, paint_r.bottom );
Canvas->LineTo( paint_r.left, paint_r.bottom );
// add extra highlight to right & bottom
Canvas->Pen->Color = clBtnShadow;
Canvas->MoveTo( paint_r.right-1, paint_r.top+1 );
Canvas->LineTo( paint_r.right-1, paint_r.bottom-1 );
Canvas->LineTo( paint_r.left+1, paint_r.bottom-1 );

// fill the background caption
Canvas->Pen->Style = psClear;
Canvas->Pen->Color = clBlack;
Canvas->Brush->Style = bsSolid;
if (fFormIsActive) Canvas->Brush->Color = clActiveCaption;
else Canvas->Brush->Color = clInactiveCaption;
if (fMicrosoftPlus) Canvas->Rectangle( paint_r.left+4, paint_r.top+4, paint_r.right-2, paint_r.top+25 );
else Canvas->Rectangle( paint_r.left+4, paint_r.top+4, paint_r.right-2, paint_r.top+23 );

// draw line between caption bar and main draw area
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnFace;
if (fMicrosoftPlus)
	{
	Canvas->MoveTo( paint_r.left+4, paint_r.top+24 );
	Canvas->LineTo( paint_r.right-3, paint_r.top+24 );
   }
else
	{
	Canvas->MoveTo( paint_r.left+4, paint_r.top+22 );
	Canvas->LineTo( paint_r.right-3, paint_r.top+22 );
   }

// draw the scrollbars
TRect scroll_r;
if (fMicrosoftPlus) scroll_r = Rect( paint_r.left+3, paint_r.top+24, paint_r.right-2, paint_r.bottom-2 );
else scroll_r = Rect( paint_r.left+3, paint_r.top+23, paint_r.right-2, paint_r.bottom-2 );

PaintScrollBars( scroll_r );

// move things a bit
paint_r.top += 6;
paint_r.right -= 5;

// if there is a systemmenu icon available
if (BorderIcons.Contains(biSystemMenu))
	{
	// paint the close window ("X") button
    paint_r.right -= fIconWidth;
    PaintCloseButton( paint_r.right, paint_r.top );
    paint_r.right -= 2;

    // if either the maximize or minimize icons are available, _must_ draw both
    if (BorderIcons.Contains(biMaximize) || BorderIcons.Contains(biMinimize))
		{
		// paint the maximize window ("_") button
	    paint_r.right -= fIconWidth;
		if (WindowState==wsMaximized) PaintRestoreButton( paint_r.right, paint_r.top );
		else PaintMaximizeButton( paint_r.right, paint_r.top );

		// paint the minimize window ("_") button
		paint_r.right -= fIconWidth;
		if (WindowState==wsMinimized) PaintRestoreButton( paint_r.right, paint_r.top );
		else PaintMinimizeButton( paint_r.right, paint_r.top );
	    }

	// draw the "command" button
	PaintCommandButton( paint_r.left+6, paint_r.top-1 );
	paint_r.left += fIconWidth;
    }

// draw the caption text
paint_r.left += 8;
if (fMicrosoftPlus)
	{
   paint_r.top --;
	paint_r.bottom = paint_r.top + 18;
   }
else
	paint_r.bottom = paint_r.top + 14;


Canvas->Brush->Style = bsClear;
Canvas->Font->Height = paint_r.bottom-paint_r.top;
Canvas->Font->Style = Canvas->Font->Style << fsBold;
Canvas->Font->Name = fCaptionFontName;
if (fFormIsActive) Canvas->Font->Color = clCaptionText;
else Canvas->Font->Color = clInactiveCaptionText;

// note: need to abbreviate caption so that if too large it truncates string, appends "..."
TRect tpaint_r = Rect( paint_r.left, paint_r.top, paint_r.right, paint_r.bottom );
Canvas->TextRect( tpaint_r, tpaint_r.Left, tpaint_r.Top, Caption );
}
//---------------------------------------------------------------------------
void 	__fastcall	TDirectDrawForm::PaintSizeToolWinFrame( RECT paint_r )
{
// modify supplied coords for pixel rounding errors
paint_r.right--;
paint_r.bottom--;

// draw basic frame
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnFace;
for (int i=0;i<4;i++)
	{
	Canvas->MoveTo( paint_r.left+i, paint_r.top+i );
	Canvas->LineTo( paint_r.right-i, paint_r.top+i );
	Canvas->LineTo( paint_r.right-i, paint_r.bottom-i );
	Canvas->LineTo( paint_r.left+i, paint_r.bottom-i );
	Canvas->LineTo( paint_r.left+i, paint_r.top+i );
    }
// highlight left & top
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnHighlight;
Canvas->MoveTo( paint_r.left+1, paint_r.bottom );
Canvas->LineTo( paint_r.left+1, paint_r.top+1 );
Canvas->LineTo( paint_r.right, paint_r.top+1 );
// darken right & bottom
Canvas->Pen->Color = cl3DDkShadow;
Canvas->MoveTo( paint_r.right, paint_r.top );
Canvas->LineTo( paint_r.right, paint_r.bottom );
Canvas->LineTo( paint_r.left, paint_r.bottom );
// add extra highlight to right & bottom
Canvas->Pen->Color = clBtnShadow;
Canvas->MoveTo( paint_r.right-1, paint_r.top+1 );
Canvas->LineTo( paint_r.right-1, paint_r.bottom-1 );
Canvas->LineTo( paint_r.left+1, paint_r.bottom-1 );

// fill the background caption
Canvas->Pen->Style = psClear;
Canvas->Pen->Color = clBlack;
Canvas->Brush->Style = bsSolid;
if (fFormIsActive) Canvas->Brush->Color = clActiveCaption;
else Canvas->Brush->Color = clInactiveCaption;
if (fMicrosoftPlus) Canvas->Rectangle( paint_r.left+4, paint_r.top+4, paint_r.right-2, paint_r.top+23 );
Canvas->Rectangle( paint_r.left+4, paint_r.top+4, paint_r.right-2, paint_r.top+20 );

// draw line between caption bar and main draw area
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnFace;
if (fMicrosoftPlus)
	{
	Canvas->MoveTo( paint_r.left+4, paint_r.top+22 );
	Canvas->LineTo( paint_r.right-3, paint_r.top+22 );
   }
else
	{
	Canvas->MoveTo( paint_r.left+4, paint_r.top+19 );
	Canvas->LineTo( paint_r.right-3, paint_r.top+19 );
	}

// draw the scrollbars
TRect scroll_r;
if (fMicrosoftPlus) scroll_r = Rect( paint_r.left+3, paint_r.top+23, paint_r.right-2, paint_r.bottom-2 );
else scroll_r = Rect( paint_r.left+3, paint_r.top+22, paint_r.right-2, paint_r.bottom-2 );
PaintScrollBars( scroll_r );

// move things a bit
paint_r.top += 5;
paint_r.right -= 4;
// if there is a systemmenu icon available
if (BorderIcons.Contains(biSystemMenu))
	{
	// paint the close window ("X") button
	paint_r.right -= fIconWidth;
    PaintCloseButton( paint_r.right, paint_r.top );
    paint_r.right -= 2;
    }

// draw the caption text
paint_r.left += 6;
if (fMicrosoftPlus) paint_r.bottom = paint_r.top + 16;
else paint_r.bottom = paint_r.top + 12;

Canvas->Brush->Style = bsClear;
Canvas->Font->Height = paint_r.bottom-paint_r.top;
if (fMicrosoftPlus) Canvas->Font->Style = Canvas->Font->Style << fsBold; // bold
else Canvas->Font->Style = Canvas->Font->Style >> fsBold;	// not bold
Canvas->Font->Name = fCaptionFontName;
if (fFormIsActive) Canvas->Font->Color = clCaptionText;
else Canvas->Font->Color = clInactiveCaptionText;

// note: need to abbreviate caption so that if too large it truncates string, appends "..."
TRect tpaint_r = Rect( paint_r.left, paint_r.top, paint_r.right, paint_r.bottom );
Canvas->TextRect( tpaint_r, tpaint_r.Left, tpaint_r.Top, Caption );
}
//---------------------------------------------------------------------------
void 	__fastcall	TDirectDrawForm::PaintToolWindowFrame( RECT paint_r )
{
// modify supplied coords for pixel rounding errors
paint_r.right--;
paint_r.bottom--;

// draw basic frame
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnFace;
for (int i=0;i<3;i++)
	{
	Canvas->MoveTo( paint_r.left+i, paint_r.top+i );
	Canvas->LineTo( paint_r.right-i, paint_r.top+i );
	Canvas->LineTo( paint_r.right-i, paint_r.bottom-i );
	Canvas->LineTo( paint_r.left+i, paint_r.bottom-i );
	Canvas->LineTo( paint_r.left+i, paint_r.top+i );
    }
// highlight left & top
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnHighlight;
Canvas->MoveTo( paint_r.left+1, paint_r.bottom );
Canvas->LineTo( paint_r.left+1, paint_r.top+1 );
Canvas->LineTo( paint_r.right, paint_r.top+1 );
// darken right & bottom
Canvas->Pen->Color = cl3DDkShadow;
Canvas->MoveTo( paint_r.right, paint_r.top );
Canvas->LineTo( paint_r.right, paint_r.bottom );
Canvas->LineTo( paint_r.left, paint_r.bottom );
// add extra highlight to right & bottom
Canvas->Pen->Color = clBtnShadow;
Canvas->MoveTo( paint_r.right-1, paint_r.top+1 );
Canvas->LineTo( paint_r.right-1, paint_r.bottom-1 );
Canvas->LineTo( paint_r.left+1, paint_r.bottom-1 );

// fill the background caption
Canvas->Pen->Style = psClear;
Canvas->Pen->Color = clBlack;
Canvas->Brush->Style = bsSolid;
if (fFormIsActive) Canvas->Brush->Color = clActiveCaption;
else Canvas->Brush->Color = clInactiveCaption;
if (fMicrosoftPlus) Canvas->Rectangle( paint_r.left+3, paint_r.top+3, paint_r.right-1, paint_r.top+22 );
else Canvas->Rectangle( paint_r.left+3, paint_r.top+3, paint_r.right-1, paint_r.top+19 );

// draw line between caption bar and main draw area
Canvas->Pen->Style = psSolid;
Canvas->Pen->Color = clBtnFace;
if (fMicrosoftPlus)
	{
	Canvas->MoveTo( paint_r.left+3, paint_r.top+21 );
	Canvas->LineTo( paint_r.right-2, paint_r.top+21 );
   }
else
	{
	Canvas->MoveTo( paint_r.left+3, paint_r.top+18 );
	Canvas->LineTo( paint_r.right-2, paint_r.top+18 );
   }

// move things a bit
paint_r.top += 4;
paint_r.right -= 3;
// if there is a systemmenu icon available
if (BorderIcons.Contains(biSystemMenu))
	{
	// paint the close window ("X") button
	paint_r.right -= fIconWidth;
    PaintCloseButton( paint_r.right, paint_r.top );
    paint_r.right -= 2;
    }

// draw the caption text
paint_r.left += 5;
if (fMicrosoftPlus) paint_r.bottom = paint_r.top + 16;
else paint_r.bottom = paint_r.top + 12;

Canvas->Brush->Style = bsClear;
Canvas->Font->Height = paint_r.bottom-paint_r.top;
if (fMicrosoftPlus) Canvas->Font->Style = Canvas->Font->Style << fsBold; // bold
else Canvas->Font->Style = Canvas->Font->Style >> fsBold;	// not bold
Canvas->Font->Name = fCaptionFontName;
if (fFormIsActive) Canvas->Font->Color = clCaptionText;
else Canvas->Font->Color = clInactiveCaptionText;

// note: need to abbreviate caption so that if too large it truncates string, appends "..."
TRect tpaint_r = Rect( paint_r.left, paint_r.top, paint_r.right, paint_r.bottom );
Canvas->TextRect( tpaint_r, tpaint_r.Left, tpaint_r.Top, Caption );
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintClientArea( HDC hdc, RECT client_r )
{
if (hdc==NULL) return;
if (WindowState==wsMinimized) return;
if (client_r.right-client_r.left<=0) return;
if (client_r.bottom-client_r.top<=0) return;

// move the draw start location to the top,left of the form's client area
MoveWindowOrg( hdc, client_r.left, client_r.top );

// clear underneath the area we need to repaint
if (!fTransparent)
   	{
	Canvas->Handle = hdc;
	Canvas->Brush->Color = Color;
	Canvas->Brush->Style = bsSolid;
	Canvas->FillRect( ClientRect );
	Canvas->Handle = NULL;
    }

// paint the contents of the form, including all controls, etc.
Messages::TWMPaint	MyPaintMessage;
MyPaintMessage.Msg = WM_PAINT;
MyPaintMessage.DC  = hdc;
MyPaintMessage.Unused = 0;
MyPaintMessage.Result = 0;
PaintHandler(MyPaintMessage);

// reset the DC's window paint origin
MoveWindowOrg( hdc, -client_r.left, -client_r.top );
}
//---------------------------------------------------------------------------
void 	__fastcall TDirectDrawForm::WndProc(Messages::TMessage &Message)
{
if (DirectDraw == NULL)
	{
    TForm::WndProc(Message);
    return;
    }

// save the current message so that we can process the message both pre-vcl & post-vcl.
int		message = Message.Msg;
int		lparam	= Message.LParam;
int		wparam	= Message.WParam;

// debug
if (DirectDraw!=NULL)
	if (!fHandlingWMPAINT)
		{
	    System::AnsiString msg = MessageStr( message, wparam, lparam );
		if (msg!="") DirectDraw->Debug(DBG_INFO,Name+"::","Event->"+msg);
	    }

// pre-VCL message handling
switch (message)
	{
    case WM_PAINT:

		// if we're using DirectDraw stuff & we're not already handling a paint message &
        // we're currently visible
    	if ((DirectDraw != NULL) && !fHandlingWMPAINT && Visible)

            	// note: TForm::WndProc generates a WM_PAINT during it's handling of
                // minimize, maximize & restore messages. Problem is that the WM_PAINT is
                // generated before TForm::WindowState is updated, causing us problems when
                // the painting reacts to the WindowState. thus we set flags just before
                // these events occur & ignore the WM_PAINT message at this point.

            	if (!fFormIsBeingMinimized && !fFormIsBeingMaximized && !fFormIsBeingRestored)
		            {
			    	// no point painting ANYTHING if there is another window that is maximized
			        for (int form=0;form<Screen->FormCount;form++)
                        if ((Screen->Forms[form] != this) &&
                        	(Screen->Forms[form]->Visible) &&
                        	(Screen->Forms[form]->WindowState == wsMaximized) &&
                            (Screen->ActiveForm != this)) //? why this is needed ?
                            {
                            Message.Result = 0;
							return;
                            }

 	                // set a flag so that we don't handle multiple WM_PAINT messages while
	                // we're still handling the first WM_PAINT message (this _does_ happen)
		            fHandlingWMPAINT = true;

			        // get a back buffer DC from DirectDraw iff possible
	                HDC hdc = DirectDraw->GetBackBufferDC();

	    			// if the DC retrieved is valid
			        if (hdc != NULL)
			    	   	{
                        // select & save the bounds of the entire window for drawing
                        RECT form_r = SelectEntireForm( hdc );

						// paint the frame
                        PaintWindowFrame( hdc, form_r );

                        // select only the client area
                        RECT client_r = SelectClientArea( hdc, form_r );

                        // paint the client area
                        PaintClientArea( hdc, client_r );

					    // release the DirectX supplied DC
						DirectDraw->ReleaseBackBufferDC();

	                    // flip (blit) the stuff from the backbuffer to the display
						fFramesFlipped++;
//                        DirectDraw->FlipBackBuffer();
		                DirectDraw->FlipBackBufferRect(form_r.left,form_r.top,form_r.right,form_r.bottom);

			            // tell windows that the entire window has been repainted
   			            ValidateRect(Handle,NULL);
			            }

	                // reset the flag used to stop multiple WM_PAINT messages
			        fHandlingWMPAINT = false;

			        // make sure windows knows we've handled the message
                    Message.Result = 0;
                    return;
                    }
    			else
                	{
                    // simply clear relevant flags
                    if (fFormIsBeingRestored) 	fFormIsBeingRestored = false;
	            	if (fFormIsBeingMinimized) 	fFormIsBeingMinimized = false;
                    if (fFormIsBeingMaximized) 	fFormIsBeingMaximized = false;
                    // tell windows that the window has been redrawn (even though it hasn't)
                    ValidateRect( Handle, NULL );
                    }
		// let the VCL handle the rest of the message
        break;

	case WM_ERASEBKGND:
    	// tell windows that we've handled the message & skip out
        Message.Result = 0;
        return;

    case WM_ACTIVATE:

		// if we're possibly using DirectDraw stuff & we're not already handling a paint message
    	if ((DirectDraw != NULL) && !fHandlingWMPAINT)
        	{
            int active_flag = LOWORD(Message.WParam);
            fFormIsActive = ((active_flag == WA_ACTIVE) || (active_flag == WA_CLICKACTIVE));
            if (fFormIsActive)
            	{
//                Perform( WM_PAINT, 0, 0 );
                Message.Result = 0;
                return;
                }
 	        }
        break;

	case WM_MENUSELECT:

		{
//        UINT uItem = (UINT) LOWORD(wparam);   // menu item or submenu index
		UINT fuFlags = (UINT) HIWORD(wparam); // menu flags
		HMENU hmenu = (HMENU) lparam;          // handle of menu clicked
		// if the fuFlags parameter contains 0xFFFF and the hmenu parameter contains NULL, Windows has closed the menu.
		if ((fuFlags == 0xFFFF) && (hmenu == NULL))
        	{
            fMenuIsDisplayed = false;
//			if (DirectDraw != NULL) DirectDraw->Debug(DBG_NOTE,"<<FORM>>::WndProc()","menu is not displayed");
            }
        else
        	{
        	fMenuIsDisplayed = true;
//			if (DirectDraw != NULL) DirectDraw->Debug(DBG_NOTE,"<<FORM>>::WndProc()","menu IS displayed");
            }
        }
		break;

	case WM_SYSCOMMAND:

    	// determine exactly what the form is _about_ to do, so that we can respond to events
        // that are seemingly generated _during_ processing of the WM_SYSCOMMAND
        {
        // mask out the actual message
        int sysmsg = (wparam & 0x0000FFF0);

        if (DirectDraw != NULL)
        	{
            if (sysmsg==SC_MOVE)
            	{
                int xpos = LOWORD(lparam);
                int ypos = HIWORD(lparam);
                DirectDraw->SetMousePos( xpos, ypos ); 
                }
            }

        // if we are the main form, we are only interested in saving the fact that the form
        // is (about to be) minimized.
        if (DirectDraw != NULL)
           	{
			if (DirectDraw->Owner == this)
               	{
		        bool minimized = ((sysmsg == SC_MINIMIZE) || (sysmsg == SC_ICON));
	            if (minimized)
	              	{
                    fMainFormIsMinimized = true;
			        fDisplayWidth = DirectDraw->Width;
			        fDisplayHeight = DirectDraw->Height;
			        fDisplayBitsPerPixel = DirectDraw->BitsPerPixel;
	                }
                else
			        if (fMainFormIsMinimized)
			           	{
			            DirectDraw->SetDisplayMode( fDisplayWidth, fDisplayHeight, fDisplayBitsPerPixel );
			            fMainFormIsMinimized = false;
			            }
                }
	        else
	        	// determine exactly what is about to happen for any other form.
		        // the reason for doing this is that when the vcl / windows handles this message, it
		        // seemingly generates a WM_PAINT message _before_ updating TForm::WindowState.
		        // otherwise we cannot accurately paint the maximize/restore buttons, etc.
	        	{
		        fFormIsBeingRestored  = (sysmsg == SC_RESTORE);
			    fFormIsBeingMinimized = ((sysmsg == SC_MINIMIZE) || (sysmsg == SC_ICON));
			    fFormIsBeingMaximized = ((sysmsg == SC_MAXIMIZE) || (sysmsg == SC_ZOOM));
	            }
            }
	    }
		break;

	case CM_FOCUSCHANGED:

        if (DirectDraw!=NULL)
        	{
			// make sure to unpaint the mouse cursor before the vcl / gdi highlights the
            // new object of focus
            DirectDraw->Debug( DBG_NOTE, Name, "before focus changes, unpainting cursor" );
		    DirectDraw->UnpaintMouseCursor();
            DirectDraw->MouseBusy = this;
            }
    	break;

	}

// let the VCL handle the message
TForm::WndProc(Message);

// now that the VCL / windows has handled the message, handle anything extra required

switch (message)
	{
    case WM_COMMAND:

        if (DirectDraw!=NULL)
        	if (DirectDraw->MouseBusy==this)
	        	{
				// make sure to repaint the mouse over the top of the control
		        DirectDraw->Debug( DBG_NOTE, Name, "after control has been painted, painting cursor" );
		        DirectDraw->MouseBusy = NULL;
			    DirectDraw->PaintMouseCursor();
	            }
        break;

    case WM_SIZE:
    case WM_MOVE:

    	if (fFormIsActive) Perform( WM_PAINT, 0, 0 );
//        if (DirectDraw != NULL) DirectDraw->SetMouseCursor( NULL );
        break;


    case WM_SETCURSOR:

 		if (DirectDraw!=NULL) DirectDraw->SetMouseCursor( Screen->Cursors[Screen->Cursor] );
//		if (DirectDraw!=NULL) DirectDraw->UpdateMouseCursor(true);
        break;
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::PaintHandler(Messages::TWMPaint &Message)
{
// simply check our preconditions & skip out if not appropriate
if (DirectDraw == NULL)
	{
    TForm::PaintHandler(Message);
    return;
    }

// if the supplied DC is okay
HDC hdc = Message.DC;
// use standard function calls to paint to our DirectX surface
try
  	{
    int	SaveIndex;
    int Clip;

    if (ControlCount==0)
    	{
//		DirectDraw->Debug(DBG_NOTE,"<<FORM>>::PaintHandler()","no controls, simply painting window");
        PaintWindow(hdc);
        }
    else
    	{
//		DirectDraw->Debug(DBG_NOTE,"<<FORM>>::PaintHandler()","has controls, calculating clip extent");
        SaveIndex = SaveDC( hdc );
	    Clip = SIMPLEREGION;
        for (int i=0;i<ControlCount;i++)
			{
            bool control_is_visible = Controls[i]->Visible;
		    bool control_design_visible = !Controls[i]->ControlStyle.Contains(csNoDesignVisible);

		   	if (control_is_visible && control_design_visible)
				{
	            Clip = ExcludeClipRect( hdc, Controls[i]->Left,
	            							 Controls[i]->Top,
	            							 Controls[i]->Left + Controls[i]->Width,
	                                         Controls[i]->Top + Controls[i]->Height );
	            if (Clip == NULLREGION) break;
	            }
            }
        if (Clip != NULLREGION)
        	{
//			DirectDraw->Debug(DBG_NOTE,"<<FORM>>::PaintHandler()","clip region is not null, painting window");
            PaintWindow( hdc );
            }
//        else
//   			DirectDraw->Debug(DBG_NOTE,"<<FORM>>::PaintHandler()","clip region is null, NOT painting window");
        RestoreDC( hdc, SaveIndex );
        }
	//	DirectDraw->Debug(DBG_NOTE,"<<FORM>>::PaintHandler()","painting controls");
    PaintControls( hdc, NULL );
	}
// unless a problem occurs
catch (...)
	{
//	DirectDraw->Debug(DBG_OUT,"<<FORM>>::PaintHandler()","VCL paint on DirectX DC - NOT SUCCESSFUL!");
	}
//DirectDraw->Debug(DBG_OUT,"<<FORM>>::PaintHandler()","");
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintWindow(HDC hdc)
{
// check prerequisites & skip out if they aren't correct
if (hdc == NULL) return;
if (DirectDraw == NULL)
	{
    TForm::PaintWindow(hdc);
    return;
    }

// call the normal TForm::OnPaint() event function
Canvas->Handle = hdc;
try
	{
    Paint();
    }
catch ( ... ) {}
Canvas->Handle = NULL;
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintControls(HDC hdc, TControl* First)
{
// simply check our preconditions & skip out if not appropriate
if (hdc == NULL) return;
if (DirectDraw == NULL)
	{
    TForm::PaintControls(hdc,First);
    return;
    }
if (ControlCount == 0) return;

// quickly scan the controls until we get to the first control, then fall through with the
// value of the first control in "i" (defaulting at zero if the 'First' parameter is NULL)
int i = 0;
int j = 0;
int controls_to_paint = ControlCount;
if (First != NULL)
	{
    controls_to_paint = 1;
	while (j<ControlCount)
	   	{
		if (Controls[j] == First) i = j;
	    j++;
	    }
    }

// loop through all of the visible controls, beginning with the control determined above.
int k = 0;
while (k<controls_to_paint)
	{
    // use mapped variable for easy typing
    TControl* control = Controls[i];

    // determine the bounds of the control
    RECT control_r;
    control_r.left = control->Left;
    control_r.top = control->Top;
    control_r.right = control_r.left + control->Width;
    control_r.bottom = control_r.top + control->Height;

    // and it's various states
//    bool control_is_framed = control->ControlStyle.Contains(csFramed);
    bool control_is_visible = control->Visible;
//	bool control_is_transparent = !control->ControlStyle.Contains(csOpaque);
    bool control_design_visible = !control->ControlStyle.Contains(csNoDesignVisible);
    bool control_rect_visible = RectVisible( hdc, &control_r );

    // if we really should be painting the control
   	if (control_is_visible && control_design_visible && control_rect_visible)
    	{
//        DirectDraw->Debug( DBG_INFO, Name + ":: ", "painting control " + control->Name );
        //
        // Note:        >> if the control to be painted is a wrapper for a standard windows control <<
        //
        // Cause:		This seems to a problem with windows internal architecture.
        //              For some of the standard controls (such as scrollbars) painting seems to have been
        //              implemented in such a way that the controls retrieve their own DC via standard
        //              gdi getDC(), etc. calls. These gdi calls _always_ return a newly created DC - not
        //              the DirectX DC that we want to draw on. Hence when we flip, parts of the controls
        // 				may be missing. hence flicker. YUK.
        //
        // Solution:	Replace the existing paint code with our own paint code
        //

        // dunno what this does
	    if (ControlState.Contains(csPaintCopy)) control->ControlState = control->ControlState >> csPaintCopy;

		// move the draw origin to the top, left of the control about to be painted
	    MoveWindowOrg( hdc, control->Left, control->Top );

        // determine the type of control, or ancestor class of some of the more generic controls,
        // and paint the relevant type of control.
        bool	isListBox = false;
        bool	isWinControl = false;
        bool	isScrollingWinControl = false;
	    TClass	theclass = control->ClassType();
		System::AnsiString 	class_name;

		while	(theclass != NULL)
			{
		    class_name = theclass->ClassName();
	        if (class_name == "TListBox")				{ isListBox = true; break; }
	        if (class_name == "TScrollingWinControl") 	{ isScrollingWinControl = true; break; }
	        if (class_name == "TWinControl") 			{ isWinControl = true; break; }
			theclass = theclass->ClassParent();
			}

// NOTE:
// How about trying TWinControl->Broadcast( WM_PAINT ) ?

        // special case
        if (isListBox)
        	{
            PaintListBox( hdc, (TListBox*) control );
            }
        else
        // scroling type of win control
		if (isScrollingWinControl)
        	{
            TScrollingWinControl* scrollingwincontrol = (TScrollingWinControl*) control;
            // note: this causes a WM_ERASEBKGND / WM_PAINT message pair to be handled by the control.
            // during this message pair, the scrollbars will be painted to the GDI surface, not the 'hdc'
            scrollingwincontrol->PaintTo(hdc,0,0);
            }
		else
        // derived from TWinControl
        if (isWinControl)
        	{
            TWinControl* wincontrol = (TWinControl*) control;
            wincontrol->PaintTo(hdc,0,0);
            }
        else
			// normal controls
		    control->Perform( WM_PAINT, (long) hdc, 0 );

		// move the draw origin back to where it was before the control was painted
	    MoveWindowOrg( hdc, -control->Left, -control->Top );

        // dunno what this does
	    if (ControlState.Contains(csPaintCopy)) control->ControlState = control->ControlState << csPaintCopy;
        }

	// choose the next control
    ++i %= ControlCount;	// wrap to 0..ControlCount-1
    k++;
    }
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDrawForm::Repaint()
{
// do nothing if the prerequisites don't match
if (DirectDraw == NULL)
	{
    TForm::Repaint();
    return;
    }

// force an update only when visible
if (Visible)
	{
    InvalidateRect(Handle,NULL,false);
    UpdateWindow(Handle);
    }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void __fastcall	TDirectDrawForm::ExcludeOtherForms( HWND Handle, HDC hdc )
{
// start at the z-factor of our current form
//DirectDraw->Debug(DBG_NOTE,Name + "::ExcludeOtherForms()","excluding");
HWND hwnd = GetWindow( Handle, GW_HWNDPREV );
// scan through all other windows all higher z-factors
while (hwnd != NULL)
	{
    // if it's not the current form (ie. the one being drawn)
    if (hwnd != Handle)
    	{
        // determine the visible state of the relevant form
        TForm* form = NULL;
        for (int j=0;j<Screen->FormCount;j++)
        	if (Screen->Forms[j]->Handle == hwnd)
            	{
                form = Screen->Forms[j];
                j = Screen->FormCount+1;
                }
        bool form_is_visible = false;
        if (form != NULL) form_is_visible = form->Visible;

        // if the form is visible, make sure not to paint over it...
        if (form_is_visible)
        	{
//	        DirectDraw->Debug(DBG_NOTE,Name+"::ExcludeOtherForms()","excluding '" + form->Name + "'");
            RECT exclude_r;
            // if the form is minimized, get the sizes of the minimized caption bar
			if (form->WindowState == wsMinimized)
            	{
	            // get the top, left of the minimized bar
	            WINDOWPLACEMENT wndplace;
	            wndplace.length = sizeof(WINDOWPLACEMENT);
	            GetWindowPlacement( form->Handle, &wndplace );

	            exclude_r.left = wndplace.ptMinPosition.x;
	            exclude_r.top  = wndplace.ptMinPosition.y;
   	            exclude_r.right = exclude_r.left + fMinimizedWidth + 2 * fBorderX;
	            exclude_r.bottom = exclude_r.top + fBorderY + fCaptionHeight + 2;

				ExcludeClipRect( hdc, exclude_r.left, exclude_r.top, exclude_r.right, exclude_r.bottom );
                }
            else
            	// exclude the entire form area
            	{
                exclude_r.left = form->Left;
                exclude_r.top = form->Top;
                exclude_r.right = exclude_r.left + form->Width;
                exclude_r.bottom = exclude_r.top + form->Height;
                
				ExcludeClipRect( hdc, exclude_r.left, exclude_r.top, exclude_r.right, exclude_r.bottom );
            	}
            }
        else
        	{
//            DirectDraw->Debug(DBG_NOTE,"::WndProc","form is not visible");
            }
        }
//    else
//    	DirectDraw->Debug(DBG_NOTE,"::WndProc","retrieved hwnd is current drawing window");

	// go to the next highest z-order
    hwnd = GetWindow( hwnd, GW_HWNDPREV );
    }
//DirectDraw->Debug(DBG_OUT,"::ExcludeOtherForms","" );
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::ExcludeMenu( HDC hdc, HMENU hmenu )
{
// skip out if the menu is not currently displayed
if (!fMenuIsDisplayed) return;
if (Screen->ActiveForm != this) return;

//DirectDraw->Debug(DBG_IN, "::ExcludeMenu","" );

// if the menu is null, get the default menu for this window
if (hmenu == NULL) hmenu = GetSystemMenu( Handle, false );

// assuming we have a valid menu
if (hmenu != NULL)
	{
//    DirectDraw->Debug(DBG_NOTE, "::ExcludeMenu","found a valid HMenu" );
    // determine the number of items in this menu
    int num_items = GetMenuItemCount( hmenu );
    RECT exclude_r;
    MENUITEMINFO item_info;
    item_info.cbSize = sizeof(MENUITEMINFO);

//    DirectDraw->Debug(DBG_NOTE, "::ExcludeMenu","menu has " +IntToStr(num_items) + " items");
    // loop through each item in turn, getting the region used by that item & excluding
    // it from the current draw region

    fExcludeMenu.Left = 100000;
    fExcludeMenu.Top = 100000;
    fExcludeMenu.Right = -1;
    fExcludeMenu.Bottom = -1;
    for (int i=0;i<num_items;i++)
    	{
//    DirectDraw->Debug(DBG_NOTE, "::ExcludeMenu","getting info for item # " + IntToStr(i));
        if (GetMenuItemRect( Handle, hmenu, i, &exclude_r ) != 0)
        	{
//    DirectDraw->Debug(DBG_NOTE, "::ExcludeMenu","retrieved & excluding menu item rect: " + IntToStr(exclude_r.left) + "," + IntToStr(exclude_r.top) + " - " + IntToStr(exclude_r.right) + "," + IntToStr(exclude_r.bottom));
            exclude_r.right += fBorderX*2;
            exclude_r.bottom += fBorderY*2;

//    DirectDraw->Debug(DBG_NOTE, "::ExcludeMenu","excluding menu item rect (modified): " + IntToStr(exclude_r.left) + "," + IntToStr(exclude_r.top) + " - " + IntToStr(exclude_r.right) + "," + IntToStr(exclude_r.bottom));
			ExcludeClipRect( hdc, exclude_r.left, exclude_r.top, exclude_r.right, exclude_r.bottom );

            // because the menu item is displayed on the GDI surface,
            // and we cannot force it to draw onto our backbuffer DC,
            // make sure to copy the menu item from the gdi surface to the backbuffer
            // before the blit() occurs

            // i comment: --- or manually paint the menu - it's much faster !!!

            if (exclude_r.left < fExcludeMenu.Left) 		fExcludeMenu.Left = exclude_r.left;
            if (exclude_r.top < fExcludeMenu.Top) 			fExcludeMenu.Top = exclude_r.top;
            if (exclude_r.right > fExcludeMenu.Right) 		fExcludeMenu.Right = exclude_r.right;
            if (exclude_r.bottom > fExcludeMenu.Bottom) 	fExcludeMenu.Bottom = exclude_r.bottom;
/*            HDC screendc =  CreateDC("DISPLAY",NULL,NULL,NULL);
            BitBlt( hdc, exclude_r.left, exclude_r.top,
            			exclude_r.right - exclude_r.left,
                        exclude_r.bottom - exclude_r.top,
            		screendc, exclude_r.left, exclude_r.top, SRCCOPY );
			DeleteDC( screendc );
*/
//            DirectDraw->FlipGDIRectToBackBuffer(exclude_r.left,exclude_r.top,exclude_r.right-exclude_r.left,exclude_r.bottom-exclude_r.top,true);

	        // if there are any sub menus attached to this item
            GetMenuItemInfo( hmenu, i, TRUE, &item_info );
            if (((item_info.fMask & MIIM_SUBMENU) != 0) && (item_info.hSubMenu != NULL))
            	// make sure to exclude the sub-menu's as well
                {
//    DirectDraw->Debug(DBG_NOTE, "::ExcludeMenu","item has sub-menu, excluding it also");
                ExcludeMenu( hdc, item_info.hSubMenu );
                }
            }
        }
    }
//DirectDraw->Debug(DBG_OUT, "::ExcludeMenu","" );
}
//---------------------------------------------------------------------------
void __fastcall	TDirectDrawForm::Paint3DButtonFrame( int left, int top, int right, int bottom )
{
Canvas->Pen->Style = psSolid;
Canvas->Pen->Width = 1;

// start at bottom,left corner
Canvas->MoveTo(left,bottom);
// draw left edge
Canvas->Pen->Color = clBtnHighlight;
Canvas->LineTo(left,top);
// draw top edge
Canvas->LineTo(right,top);
// draw right edge
Canvas->Pen->Color = cl3DDkShadow;
Canvas->LineTo(right,bottom);
// draw bottom edge
Canvas->LineTo(left-1,bottom);
// extra highlight
left++;
right--;
top++;
bottom--;
// draw right edge
Canvas->Pen->Color = clBtnShadow;
Canvas->MoveTo(right,top);
Canvas->LineTo(right,bottom);
// draw bottom edge
Canvas->LineTo(left-1,bottom);

Canvas->Pen->Style = psClear;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::Paint3DThumbnail( int left, int top, int right, int bottom )
{
Canvas->Pen->Style = psSolid;
Canvas->Pen->Width = 1;

// fill rect
Canvas->Pen->Style = psClear;
Canvas->Brush->Style = bsSolid;
Canvas->Brush->Color = clBtnFace;
Canvas->Rectangle( left, top, right, bottom );

// start at bottom,left corner
Canvas->Pen->Style = psSolid;
Canvas->MoveTo(left,bottom);
// draw left edge
Canvas->Pen->Color = clBtnFace;
Canvas->LineTo(left,top);
// draw top edge
Canvas->LineTo(right,top);
// draw right edge
Canvas->Pen->Color = cl3DDkShadow;
Canvas->LineTo(right,bottom);
// draw bottom edge
Canvas->LineTo(left-1,bottom);
// extra highlight
left++;
right--;
top++;
bottom--;
// draw left edge
Canvas->Pen->Color = clBtnHighlight;
Canvas->MoveTo(left, bottom);
Canvas->LineTo(left, top);
// draw top edge
Canvas->LineTo(right, top);
// draw right edge
Canvas->Pen->Color = clBtnShadow;
Canvas->MoveTo(right,top);
Canvas->LineTo(right,bottom);
// draw bottom edge
Canvas->LineTo(left-1,bottom);
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintCloseButton( int left, int top )
{
TRect button_r;
button_r.Top = top;
button_r.Left = left;
button_r.Right = left + fButtonWidth - 1;
button_r.Bottom = top + fButtonHeight - 3;

if ((BorderStyle==bsSizeToolWin) || (BorderStyle==bsToolWindow))
	{
   if (fMicrosoftPlus)
   	{
	    button_r.Top++;
	    button_r.Left++;
	    button_r.Right--;
	    button_r.Bottom--;
       }
   else
   	{
       button_r.Top++;
       button_r.Left+=4;
       button_r.Right--;
       button_r.Bottom-=2;
       }
    }

Canvas->Brush->Color = clBtnFace;
Canvas->Pen->Style = psClear;
Canvas->Rectangle( button_r.Left, button_r.Top, button_r.Right+1, button_r.Bottom+1 );

Paint3DButtonFrame( button_r.Left, button_r.Top, button_r.Right, button_r.Bottom);

// draw the crosses depending on the borderstyle used
Canvas->Pen->Color = cl3DDkShadow;
Canvas->Pen->Width = 1;
Canvas->Pen->Style = psSolid;
if ((BorderStyle==bsSizeToolWin) || (BorderStyle==bsToolWindow))
	{
   if (fMicrosoftPlus)
   	{
		// top-left to bottom-right
		// right ie. \X
		Canvas->MoveTo( button_r.Left + 5, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 12, button_r.Top + 10 );
		// left ie. X\	- note: don't move this code before above code -> wierd !
		Canvas->MoveTo( button_r.Left + 4, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 11, button_r.Top + 10 );

		// top-right to bottom-left
		// left ie. X/
		Canvas->MoveTo( button_r.Left + 10, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 3, button_r.Top + 10 );
		// right ie. /X
		Canvas->MoveTo( button_r.Left + 11, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 4, button_r.Top + 10 );
       }
   else
   	{
		// top-left to bottom-right
		// right ie. \X
		Canvas->MoveTo( button_r.Left + 3, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 8, button_r.Top + 8 );
		// left ie. X\	- note: don't move this code before above code -> wierd !
		Canvas->MoveTo( button_r.Left + 2, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 7, button_r.Top + 8 );

		// top-right to bottom-left
		// left ie. X/
		Canvas->MoveTo( button_r.Left + 6, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 1, button_r.Top + 8 );
		// right ie. /X
		Canvas->MoveTo( button_r.Left + 7, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 2, button_r.Top + 8 );
       }
   }
else
	// all other border styles
	{
   if (fMicrosoftPlus)
   	{
		// top-left to bottom-right
		// centre ie. \X\
		Canvas->MoveTo( button_r.Left + 5, button_r.Top + 4 );	// note: reqd. - dunno why
		Canvas->MoveTo( button_r.Left + 5, button_r.Top + 4 );
		Canvas->LineTo( button_r.Left + 13, button_r.Top + 12 );
		// left of centre ie. X\\ - note: don't move this code before above code -> wierd result!
		Canvas->MoveTo( button_r.Left + 5, button_r.Top + 5 );
		Canvas->LineTo( button_r.Left + 12, button_r.Top + 12 );
		// right of centre ie. \\X
		Canvas->MoveTo( button_r.Left + 6, button_r.Top + 4 );
		Canvas->LineTo( button_r.Left + 13, button_r.Top + 11 );

		// top-right to bottom-left
		// centre ie. /X/
		Canvas->MoveTo( button_r.Left + 12, button_r.Top + 4 );
		Canvas->LineTo( button_r.Left + 4, button_r.Top + 12 );
		// left of centre ie. X// - note: don't move this code before above code -> wierd !
		Canvas->MoveTo( button_r.Left + 11, button_r.Top + 4 );
		Canvas->LineTo( button_r.Left + 4, button_r.Top + 11 );
		// right of centre ie. //X
		Canvas->MoveTo( button_r.Left + 12, button_r.Top + 5 );
		Canvas->LineTo( button_r.Left + 5, button_r.Top + 12 );
       }
   else
   	{
		// top-left to bottom-right
		// right ie. \X
		Canvas->MoveTo( button_r.Left + 5, button_r.Top + 3 );	// note: reqd. - dunno why
       Canvas->MoveTo( button_r.Left + 5, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 12, button_r.Top + 10 );
		// left ie. X\ - note: don't move this code before above code -> wierd result!
		Canvas->MoveTo( button_r.Left + 4, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 11, button_r.Top + 10 );

		// top-right to bottom-left
		// right ie. /X
		Canvas->MoveTo( button_r.Left + 11, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 3, button_r.Top + 10 );
		// left ie. X/ - note: don't move this code before above code -> wierd !
		Canvas->MoveTo( button_r.Left + 10, button_r.Top + 3 );
		Canvas->LineTo( button_r.Left + 4, button_r.Top + 10 );
       }
	}
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintCommandButton( int left, int top )
{
if (!fSmallIcon) return;
fSmallIcon->Draw( Canvas, left, top, 0 );
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintRestoreButton( int left, int top )
{
TRect button_r;
button_r.Top = top;
button_r.Left = left;
button_r.Right = left + fButtonWidth - 1;
button_r.Bottom = top + fButtonHeight - 3;

Canvas->Brush->Color = clBtnFace;
Canvas->Pen->Style = psClear;
Canvas->Rectangle( button_r.Left, button_r.Top, button_r.Right+1, button_r.Bottom+1 );

Paint3DButtonFrame( button_r.Left, button_r.Top, button_r.Right, button_r.Bottom);

bool restore_active = (BorderIcons.Contains(biMaximize));
if (restore_active) Canvas->Pen->Color = cl3DDkShadow;
else Canvas->Pen->Color = clBtnShadow;
Canvas->Pen->Width = 1;
Canvas->Pen->Style = psSolid;

// top-middle to top-right
Canvas->MoveTo( button_r.Left + 6, button_r.Top + 3 );
Canvas->LineTo( button_r.Left + 13, button_r.Top + 3 );
Canvas->MoveTo( button_r.Left + 6, button_r.Top + 4 );
Canvas->LineTo( button_r.Left + 13, button_r.Top + 4 );
// middle-left to middle-middle
Canvas->MoveTo( button_r.Left + 4, button_r.Top + 6 );
Canvas->LineTo( button_r.Left + 11, button_r.Top + 6 );
Canvas->MoveTo( button_r.Left + 4, button_r.Top + 7 );
Canvas->LineTo( button_r.Left + 11, button_r.Top + 7 );
// bottom-left box
Canvas->MoveTo( button_r.Left + 4, button_r.Top + 8 );
Canvas->LineTo( button_r.Left + 4, button_r.Top + 11 );
Canvas->MoveTo( button_r.Left + 4, button_r.Top + 11 );
Canvas->LineTo( button_r.Left + 10, button_r.Top + 11 );
Canvas->MoveTo( button_r.Left + 10, button_r.Top + 8 );
Canvas->LineTo( button_r.Left + 10, button_r.Top + 12 );
// top-right box
Canvas->MoveTo( button_r.Left + 6, button_r.Top + 5 );
Canvas->LineTo( button_r.Left + 6, button_r.Top + 7 );
Canvas->MoveTo( button_r.Left + 12, button_r.Top + 5 );
Canvas->LineTo( button_r.Left + 12, button_r.Top + 9 );
Canvas->MoveTo( button_r.Left + 10, button_r.Top + 8 );
Canvas->LineTo( button_r.Left + 13, button_r.Top + 8 );
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintMinimizeButton( int left, int top )
{
TRect button_r;
button_r.Top = top;
button_r.Left = left;
button_r.Right = left + fButtonWidth - 1;
button_r.Bottom = top + fButtonHeight - 3;

Canvas->Brush->Color = clBtnFace;
Canvas->Pen->Style = psClear;
Canvas->Rectangle( button_r.Left, button_r.Top, button_r.Right+1, button_r.Bottom+1 );

Paint3DButtonFrame( button_r.Left, button_r.Top, button_r.Right, button_r.Bottom);

bool minimize_active = (BorderIcons.Contains(biMinimize));
if (minimize_active) Canvas->Pen->Color = cl3DDkShadow;
else Canvas->Pen->Color = clBtnShadow;
Canvas->Pen->Width = 1;
Canvas->Pen->Style = psSolid;

if (fMicrosoftPlus)
	{
	// top-left to top-right
	Canvas->MoveTo( button_r.Left + 5, button_r.Top + 10 );
	Canvas->LineTo( button_r.Left + 11, button_r.Top + 10 );
	Canvas->MoveTo( button_r.Left + 5, button_r.Top + 11 );
	Canvas->LineTo( button_r.Left + 11, button_r.Top + 11 );

	if (!minimize_active)
		{
	    Canvas->Pen->Color = clBtnHighlight;
	    Canvas->MoveTo( button_r.Left + 6, button_r.Top + 12 );
	    Canvas->LineTo( button_r.Left + 13, button_r.Top + 12 );
	    Canvas->MoveTo( button_r.Left + 13, button_r.Top + 9 );	// vertical highlight
	    Canvas->MoveTo( button_r.Left + 13, button_r.Top + 12 );
	    }
   }
else
	{
	// top-left to top-right
	Canvas->MoveTo( button_r.Left + 4, button_r.Top + 9 );
	Canvas->LineTo( button_r.Left + 10, button_r.Top + 9 );
	Canvas->MoveTo( button_r.Left + 4, button_r.Top + 10 );
	Canvas->LineTo( button_r.Left + 10, button_r.Top + 10 );

	if (!minimize_active)
		{
	    Canvas->Pen->Color = clBtnHighlight;
	    Canvas->MoveTo( button_r.Left + 5, button_r.Top + 11 );
	    Canvas->LineTo( button_r.Left + 12, button_r.Top + 11 );
	    Canvas->MoveTo( button_r.Left + 12, button_r.Top + 8 );	// vertical highlight
	    Canvas->MoveTo( button_r.Left + 12, button_r.Top + 11 );
	    }
   }
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintMaximizeButton( int left, int top )
{
TRect button_r;
button_r.Top = top;
button_r.Left = left;
button_r.Right = left + fButtonWidth - 1;
button_r.Bottom = top + fButtonHeight - 3;

Canvas->Brush->Color = clBtnFace;
Canvas->Pen->Style = psClear;
Canvas->Rectangle( button_r.Left, button_r.Top, button_r.Right+1, button_r.Bottom+1 );

Paint3DButtonFrame( button_r.Left, button_r.Top, button_r.Right, button_r.Bottom);

bool maximize_active = (BorderIcons.Contains(biMaximize));
if (maximize_active) Canvas->Pen->Color = cl3DDkShadow;
else Canvas->Pen->Color = clBtnShadow;

Canvas->Pen->Width = 1;
Canvas->Pen->Style = psSolid;

if (fMicrosoftPlus)
	{
	// top-left to top-right
	Canvas->MoveTo( button_r.Left + 3, button_r.Top + 3 );
	Canvas->LineTo( button_r.Left + 13, button_r.Top + 3 );
	Canvas->MoveTo( button_r.Left + 3, button_r.Top + 4 );
	Canvas->LineTo( button_r.Left + 13, button_r.Top + 4 );
	// left
	Canvas->MoveTo( button_r.Left + 3, button_r.Top + 5 );
	Canvas->LineTo( button_r.Left + 3, button_r.Top + 11 );
	// right
	Canvas->MoveTo( button_r.Left + 12, button_r.Top + 5 );
	Canvas->LineTo( button_r.Left + 12, button_r.Top + 11 );
	// bottom
	Canvas->MoveTo( button_r.Left + 3, button_r.Top + 11 );
	Canvas->LineTo( button_r.Left + 13, button_r.Top + 11 );

	// add highlights to icon when it's not available
	if (!maximize_active)
	  	{
		Canvas->Pen->Color = clBtnHighlight;
	    // top
	    Canvas->MoveTo( button_r.Left + 4, button_r.Top + 5 );
	    Canvas->LineTo( button_r.Left + 12, button_r.Top + 5 );
	    // left
	    Canvas->MoveTo( button_r.Left + 4, button_r.Top + 5 );
	    Canvas->LineTo( button_r.Left + 4, button_r.Top + 11 );
	    // right
	    Canvas->MoveTo( button_r.Left + 13, button_r.Top + 4 );
	    Canvas->LineTo( button_r.Left + 13, button_r.Top + 12 );
	    // bottom
	    Canvas->MoveTo( button_r.Left + 4, button_r.Top + 12 );
	    Canvas->LineTo( button_r.Left + 13, button_r.Top + 12 );
	    }
   }
else
	{
	// top-left to top-right
	Canvas->MoveTo( button_r.Left + 3, button_r.Top + 2 );
	Canvas->LineTo( button_r.Left + 12, button_r.Top + 2 );
	Canvas->MoveTo( button_r.Left + 3, button_r.Top + 3 );
	Canvas->LineTo( button_r.Left + 12, button_r.Top + 3 );
	// left
	Canvas->MoveTo( button_r.Left + 3, button_r.Top + 4 );
	Canvas->LineTo( button_r.Left + 3, button_r.Top + 10 );
	// right
	Canvas->MoveTo( button_r.Left + 11, button_r.Top + 4 );
	Canvas->LineTo( button_r.Left + 11, button_r.Top + 10 );
	// bottom
	Canvas->MoveTo( button_r.Left + 3, button_r.Top + 10 );
	Canvas->LineTo( button_r.Left + 12, button_r.Top + 10 );

	// add highlights to icon when it's not available
	if (!maximize_active)
	  	{
		Canvas->Pen->Color = clBtnHighlight;
	    // top
	    Canvas->MoveTo( button_r.Left + 4, button_r.Top + 4 );
	    Canvas->LineTo( button_r.Left + 11, button_r.Top + 4 );
	    // left
	    Canvas->MoveTo( button_r.Left + 4, button_r.Top + 4 );
	    Canvas->LineTo( button_r.Left + 4, button_r.Top + 10 );
	    // right
	    Canvas->MoveTo( button_r.Left + 12, button_r.Top + 4 );
	    Canvas->LineTo( button_r.Left + 12, button_r.Top + 11 );
	    // bottom
	    Canvas->MoveTo( button_r.Left + 4, button_r.Top + 11 );
	    Canvas->LineTo( button_r.Left + 12, button_r.Top + 11 );
	    }
   }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void	__fastcall TDirectDrawForm::PaintScrollBars( TRect client_r )
{
// if the vertical bar is available
if (VertScrollBar!=NULL)
    // and if the vertical bar is also visible
	if (VertScrollBar->Visible)
		// and there is enough room to paint the scrollbar
	    if (client_r.Right - client_r.Left > fScrollbarWidth)
        	// then paint the scrollbar
			PaintVScrollBar( client_r );

// if the horiz bar is available
if (HorzScrollBar!=NULL)
    // and if the horiz bar is also visible
	if (HorzScrollBar->Visible)
		// and there is enough room to paint the scrollbar
	    if (client_r.Bottom - client_r.Top > fScrollbarHeight)
        	// then paint the scrollbar
			PaintHScrollBar( client_r );

// draw the resizing control /// thingo
if (fVScrollBarIsVisible && fHScrollBarIsVisible)
	{
    // make it blank
    int top = client_r.Bottom-fScrollbarHeight-1;
    int left = client_r.Right-fScrollbarWidth-1;
    Canvas->Pen->Color = clBtnFace;
    Canvas->Pen->Style = psSolid;
    Canvas->Brush->Color = clBtnFace;
    Canvas->Brush->Style = bsSolid;
    Canvas->Rectangle( left, top, left+fScrollbarWidth, top+fScrollbarHeight );
    // add the lowlights
    Canvas->Pen->Color = clInactiveCaption;
    Canvas->MoveTo( left + 5, top + 15 );
    Canvas->LineTo( left + 16, top + 4 );
    Canvas->MoveTo( left + 6, top + 15 );
    Canvas->LineTo( left + 16, top + 5 );
    Canvas->MoveTo( left + 9, top + 15 );
    Canvas->LineTo( left + 16, top + 8 );
    Canvas->MoveTo( left + 10, top + 15 );
    Canvas->LineTo( left + 16, top + 9 );
    Canvas->MoveTo( left + 13, top + 15 );
    Canvas->LineTo( left + 16, top + 12 );
    Canvas->MoveTo( left + 14, top + 15 );
    Canvas->LineTo( left + 16, top + 13 );
    // add the highlights
    Canvas->Pen->Color = clBtnHighlight;
    Canvas->MoveTo( left + 4, top + 15 );
    Canvas->LineTo( left + 16, top + 3 );
    Canvas->MoveTo( left + 8, top + 15 );
    Canvas->LineTo( left + 16, top + 7 );
    Canvas->MoveTo( left + 12, top + 15 );
    Canvas->LineTo( left + 16, top + 11 );
    }
}
//---------------------------------------------------------------------------
void 	__fastcall	TDirectDrawForm::PaintVScrollBar( TRect client_r )
{
// skip out if there's not scrollbar to paint
if (VertScrollBar==NULL) return;

// icon's are slightly smaller than standard, use this smaller value
int iconheight = fIconHeight-2;

// determine the bounds of the entire scroll bar
int top = client_r.Top;
int left = client_r.Right - fScrollbarWidth - 1;
int right = client_r.Right - 1;
int bottom = client_r.Bottom - 1;

// if the horizontal scroll bar is currently visible, reduce the available area
if (fHScrollBarIsVisible) bottom -= fScrollbarHeight;

// determine the tops of the various bits
int topofuparrow = top;
int topofdnarrow = bottom - iconheight;
int topofwhitespace = topofuparrow + iconheight;

// the amount of the form that is visible, converted to pixels
int clientheight = bottom-top;

// height of the "whitespace" between the arrows
int whitespaceheight = topofdnarrow-topofwhitespace-1;

// --- calculate the height and start pixel position of the thumbnail
// note: 	the following params are supplied by windows :-
//				range == maximum size of the client area when fully exposed
//				pos == anywhere between 0..maxpos
//				where (calculated) maxpos == (supplied) range - (calculated) clientheight;
//				and (calculated) clientheight == the height of the client area (from supplied paint rect)
int pos = VertScrollBar->Position;
int range = VertScrollBar->Range;
int maxpos = range - clientheight;

int minimum_thumbnail_height = 8;
int thumbnailheight = minimum_thumbnail_height;
int topofthumbnail = topofwhitespace;

if (range!=0) thumbnailheight = whitespaceheight * clientheight / range;
if (thumbnailheight<minimum_thumbnail_height) thumbnailheight=minimum_thumbnail_height;
if (maxpos!=0) topofthumbnail += (whitespaceheight-thumbnailheight) * pos / maxpos;

// --- draw the scrollbar only when required

// (1) paint a modified scrollbar if the area has been reduced until it is smaller than the thumbnail
if (whitespaceheight<minimum_thumbnail_height)
	{
    // the scrollbar is diplayed
    fVScrollBarIsVisible = true;

    // if there is any whitespace at all,
    if (whitespaceheight>0)
    	{
		// draw up arrow
		Canvas->Draw( left, topofuparrow, fUpArrow );

		// draw down arrow
		Canvas->Draw( left, topofdnarrow, fDnArrow );

		// draw under the thumbnail
		Canvas->Pen->Style = psClear;
		Canvas->Brush->Style = bsSolid;
		Canvas->Brush->Color = clScrollBar;
		Canvas->Rectangle( left, topofwhitespace, right+1, topofdnarrow+1 );
        }
    else
    	// stretch draw the arrow icons so that they fit the available area
    	{
        int half_height = (bottom-top) / 2;
        TRect UpArrow_r = Rect( left, topofuparrow, right, topofuparrow+half_height );
        TRect DnArrow_r = Rect( left, UpArrow_r.Bottom, right, bottom );

		// draw up arrow
		Canvas->StretchDraw( UpArrow_r, fUpArrow );

		// draw down arrow
		Canvas->StretchDraw( DnArrow_r, fDnArrow );
        }
    }
else
	// (2) paint the scrollbar if the client area is not fully displayed
	if (thumbnailheight<=whitespaceheight)
    	{
	    // the scrollbar is diplayed
	    fVScrollBarIsVisible = true;

		// draw up arrow
		Canvas->Draw( left, topofuparrow, fUpArrow );

		// draw down arrow
		Canvas->Draw( left, topofdnarrow, fDnArrow );

		// draw under the thumbnail
		Canvas->Pen->Style = psClear;
		Canvas->Brush->Style = bsSolid;
		Canvas->Brush->Color = clScrollBar;
		Canvas->Rectangle( left, topofwhitespace, right+1, topofdnarrow+1 );

		// draw the thumbnail
	    Paint3DThumbnail( left, topofthumbnail, right-1, topofthumbnail+thumbnailheight );
        }
    else
        // (3) don't bother painting the scrollbar at all
		fVScrollBarIsVisible = false;
}
//---------------------------------------------------------------------------
void 	__fastcall	TDirectDrawForm::PaintHScrollBar( TRect client_r )
{
// skip out if there's not scrollbar to paint
if (HorzScrollBar==NULL) return;

// icon's are slightly smaller than standard, use this smaller value
int iconwidth = fIconWidth-2;

// determine the bounds of the entire scroll bar
int top = client_r.Bottom - fScrollbarHeight - 1;
int left = client_r.Left + 1;		 // note: left + 1 !!
int right = client_r.Right - 1;
int bottom = client_r.Bottom - 1;

// if the vertical scroll bar is currently visible, reduce the available area
if (fVScrollBarIsVisible) right -= fScrollbarWidth;

// determine the lefts of the various bits
int leftofltarrow = left;
int leftofrtarrow = right - iconwidth;
int leftofwhitespace = leftofltarrow + iconwidth;

// the amount of the form that is visible, converted to pixels
int clientwidth = right-left;

// width of the "whitespace" between the arrows
int whitespacewidth = leftofrtarrow-leftofwhitespace-1;

// --- calculate the width and start pixel position of the thumbnail
// note: 	the following params are supplied by windows :-
//				range == maximum size of the client area when fully exposed
//				pos == anywhere between 0..maxpos
//				where (calculated) maxpos == (supplied) range - (calculated) clientwidth;
//				and (calculated) clientwidth == the width of the client area (from supplied paint rect)
int pos = HorzScrollBar->Position;
int range = HorzScrollBar->Range;
int maxpos = range - clientwidth;

int minimum_thumbnail_width = 8;
int thumbnailwidth = minimum_thumbnail_width;
int leftofthumbnail = leftofwhitespace;

if (range!=0) thumbnailwidth = whitespacewidth * clientwidth / range;
if (thumbnailwidth<minimum_thumbnail_width) thumbnailwidth=minimum_thumbnail_width;
if (maxpos!=0) leftofthumbnail += (whitespacewidth-thumbnailwidth) * pos / maxpos;

// --- draw the scrollbar only when required

// (1) paint a modified scrollbar if the area has been reduced until it is smaller than the thumbnail
if (whitespacewidth<minimum_thumbnail_width)
	{
    // the scrollbar is diplayed
    fHScrollBarIsVisible = true;

    // if there is any whitespace at all,
    if (whitespacewidth>0)
    	{
		// draw lt arrow
		Canvas->Draw( leftofltarrow, top, fLtArrow );

		// draw right arrow
		Canvas->Draw( leftofrtarrow, top, fRtArrow );

		// draw under the thumbnail
		Canvas->Pen->Style = psClear;
		Canvas->Brush->Style = bsSolid;
		Canvas->Brush->Color = clScrollBar;
		Canvas->Rectangle( leftofwhitespace, top, leftofrtarrow+1, bottom+1 );
        }
    else
    	// stretch draw the arrow icons so that they fit the available area
    	{
        int half_width = (right-left) / 2;
        TRect LtArrow_r = Rect( leftofltarrow, top, leftofltarrow+half_width, bottom );
        TRect RtArrow_r = Rect( LtArrow_r.Right, top, right, bottom );

		// draw lt arrow
		Canvas->StretchDraw( LtArrow_r, fLtArrow );

		// draw right arrow
		Canvas->StretchDraw( RtArrow_r, fRtArrow );
        }
    }
else
	// (2) paint the scrollbar if the client area is not fully displayed
	if (thumbnailwidth<=whitespacewidth)
    	{
	    // the scrollbar is diplayed
	    fHScrollBarIsVisible = true;

		// draw lt arrow
		Canvas->Draw( leftofltarrow, top, fLtArrow );

		// draw right arrow
		Canvas->Draw( leftofrtarrow, top, fRtArrow );

		// draw under the thumbnail
		Canvas->Pen->Style = psClear;
		Canvas->Brush->Style = bsSolid;
		Canvas->Brush->Color = clScrollBar;
		Canvas->Rectangle( leftofwhitespace, top, leftofrtarrow+1, bottom+1 );

		// draw the thumbnail
		Paint3DThumbnail( leftofthumbnail, top, leftofthumbnail+thumbnailwidth, bottom-1 );
        }
    else
        // (3) don't bother painting the scrollbar at all
		fHScrollBarIsVisible = false;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintListBox( HDC hdc, TListBox* listbox )
{
// NOTE: when using the standard PaintTo() method to paint the control, most of the
// control gets painted to the supplied DC, but the scrollbars & cursor aren't painted.
// This has something to do with the way they (microsoft programmers) have optimized
// drawing once upon a time - there is only ONE display device ever, and only elements that
// have changed need to be repainted, never the whole scrollbar or whatever.
// MEANING: we have to manually paint the scrollbars, etc. Yuk, but that's life.

listbox->Canvas->Handle = hdc;
listbox->PaintTo(hdc,0,0);
listbox->Canvas->Handle = NULL;

// get information re: the vertical scrollbar
SCROLLINFO	scrollinfo;
scrollinfo.cbSize = sizeof(SCROLLINFO);
scrollinfo.fMask = SIF_ALL;
bool info_retrieved = (GetScrollInfo( listbox->Handle, SB_VERT, &scrollinfo ) != 0);
int rowsperpage = scrollinfo.nPage;
int maxrows = rowsperpage;
if (!listbox->IntegralHeight) maxrows++;

// if the info retrieved was valid and we need to display a scrollbar
if (info_retrieved && (listbox->Items->Count>maxrows))
	{
    Canvas->Handle = hdc;

	// vertical scroll bar
	int scrollbarpos = GetScrollPos( listbox->Handle, SB_VERT );
	int rowoffset = listbox->ItemIndex;
    int width = listbox->Width;
	int height = listbox->Height;
	int top = 2;
	int left = width - fScrollbarWidth - 2;
	int rowheight = listbox->ItemHeight;
    int totalrows = scrollinfo.nMax - scrollinfo.nMin + 1;

    // paint the vertical scrollbar
    PaintControlVScrollBar( Canvas->Handle, listbox->Handle, left, top, width, height, scrollbarpos, rowoffset, rowsperpage, totalrows );

	// now paint the windows "caret"
    // ie. draw the currently highlighted selections iff required
    if (listbox->ItemIndex != -1)
    	{
        // draw the rows of displayed text
        Canvas->Pen->Style = psClear;
		Canvas->Brush->Style = bsSolid;

        // scan through the number of visible rows
		for (int row=0;row<maxrows;row++)
        	{
            int tempindex = scrollbarpos + row;
            if (tempindex>listbox->Items->Count-1) break;

	        // draw the text for the highlight
	        TRect r;
	        r.Left = 2;
	        r.Right = left;
	        r.Top = row * rowheight + 2;
	        r.Bottom = r.Top + rowheight;
            if (r.Top>=listbox->Height-7) break;
            if (r.Bottom-r.Top>=listbox->Height-7) r.Bottom = r.Top + listbox->Height-7;
	        Canvas->Font->Name = listbox->Font->Name;
	        Canvas->Font->Style = listbox->Font->Style >> fsBold >> fsItalic >> fsUnderline;
	        Canvas->Font->Size = listbox->Font->Size;
			if (listbox->Selected[tempindex])
            	{
                Canvas->Font->Color = clWindow;
                Canvas->Brush->Color = clActiveCaption;
		        Canvas->TextRect( r, r.Left+2, r.Top, listbox->Items->Strings[tempindex] );
                }
            else
            	{
                Canvas->Font->Color = listbox->Font->Color;
                Canvas->Brush->Color = listbox->Color;
		        Canvas->TextRect( r, r.Left+2, r.Top, listbox->Items->Strings[tempindex] );
                }

            // draw the current row highlight as well
            if (listbox->Selected[tempindex] && (listbox->ItemIndex == tempindex) && (Screen->ActiveControl==listbox))
            	{
		        int i;
		        int even = listbox->ItemIndex % 2;
	            int rtop = r.Top;
	            int rleft = r.Left;
	            int rright = r.Right-1;
	            int rbottom = r.Bottom-1;

	            if (even)
                	{
			        for (i=rleft;i<rright;i+=2)
	                	{
		                Canvas->Pixels[i][rtop] = clWindow;
	                    Canvas->Pixels[i][rbottom] = clWindow;
	   	                }
                    for (i=rtop;i<rbottom;i+=2)
	                	{
		                Canvas->Pixels[rleft][i] = clWindow;
	                    Canvas->Pixels[rright][i] = clWindow;
                        }
                    Canvas->Pixels[rright][rbottom] = clWindow;
                    }
                else
	            	{
			        for (i=rleft+1;i<rright;i+=2)
	                	{
		                Canvas->Pixels[i][rtop] = clWindow;
	                    Canvas->Pixels[i][rbottom] = clWindow;
	   	                }
                    for (i=rtop+1;i<rbottom;i+=2)
	                	{
		                Canvas->Pixels[rleft][i] = clWindow;
	                    Canvas->Pixels[rright][i] = clWindow;
                    	}
                    }
                }
            }
        }
    Canvas->Handle = NULL;
    }
}
//---------------------------------------------------------------------------
void __fastcall	TDirectDrawForm::PaintControlVScrollBar( HDC hdc, HWND hwnd, int left, int top, int width, int height, int pos, int rowoffset, int rowsperpage, int totalrows )
{
// figure some simple values
int topofuparrow = top;
int bottomofuparrow = top + fIconHeight;
int topofdnarrow = height - fScrollbarHeight - 2;
int bottomofdnarrow = height - 2;
int topofthumbnailarea = bottomofuparrow-2;
int bottomofthumbnailarea = topofdnarrow+1;

// draw under the thumbnail
Canvas->Pen->Style = psClear;
Canvas->Pen->Color = clScrollBar;
Canvas->Brush->Style = bsSolid;
Canvas->Brush->Color = clScrollBar;
Canvas->Rectangle( left, topofthumbnailarea, left + fIconWidth - 1, bottomofthumbnailarea );

// draw up arrow
Canvas->Draw( left, topofuparrow, fUpArrow );

// draw down arrow
Canvas->Draw( left, topofdnarrow, fDnArrow );

// draw the thumbnail
int pixelheight = bottomofthumbnailarea - topofthumbnailarea;
int topofthumbnail = topofthumbnailarea;
int thumbnailheight = 0;
if (totalrows != 0)
   	{
    thumbnailheight += (int)(((float)pixelheight * (float)rowsperpage) / (float)totalrows);
    topofthumbnail += (int)(((float)pixelheight * (float)pos) / (float)totalrows);
    if (pos+rowsperpage >= totalrows) topofthumbnail--;
    }
if (thumbnailheight<8) thumbnailheight=8;

// frame the thumbnail
Paint3DThumbnail( left, topofthumbnail, left + fIconWidth - 3, topofthumbnail + thumbnailheight );
}
//---------------------------------------------------------------------------
void __fastcall	TDirectDrawForm::PaintControlHScrollBar( HDC hdc, HWND hwnd, int left, int top, int width, int height, int pos, int rowoffset, int rowsperpage, int totalrows )
{
// figure some simple values
int topofuparrow = top;
int bottomofuparrow = top + fIconHeight;
int topofdnarrow = height - fScrollbarHeight - 2;
int bottomofdnarrow = height - 2;
int topofthumbnailarea = bottomofuparrow-2;
int bottomofthumbnailarea = topofdnarrow+1;

// draw under the thumbnail
Canvas->Pen->Style = psClear;
Canvas->Pen->Color = clScrollBar;
Canvas->Brush->Style = bsSolid;
Canvas->Brush->Color = clScrollBar;
Canvas->Rectangle( left, topofthumbnailarea, left + fIconWidth - 1, bottomofthumbnailarea );

// draw up arrow
Canvas->Draw( left, topofuparrow, fUpArrow );

// draw down arrow
Canvas->Draw( left, topofdnarrow, fDnArrow );

// draw the thumbnail
int pixelheight = bottomofthumbnailarea - topofthumbnailarea;
int topofthumbnail = topofthumbnailarea;
int thumbnailheight = 0;
if (totalrows != 0)
   	{
    thumbnailheight += (int)(((float)pixelheight * (float)rowsperpage) / (float)totalrows);
    topofthumbnail += (int)(((float)pixelheight * (float)pos) / (float)totalrows);
    if (pos+rowsperpage >= totalrows) topofthumbnail--;
    }
if (thumbnailheight<8) thumbnailheight=8;

// frame the thumbnail
Paint3DThumbnail( left, topofthumbnail, left + fIconWidth - 3, topofthumbnail + thumbnailheight-1 );
}
//---------------------------------------------------------------------------
void __fastcall	TDirectDrawForm::BeforeMove(TObject* Sender)
{
}
//---------------------------------------------------------------------------
void __fastcall	TDirectDrawForm::AfterMove(TObject* Sender)
{
}
//---------------------------------------------------------------------------
System::AnsiString	__fastcall	TDirectDrawForm::MessageStr( int message, int wparam, int lparam )
{
System::AnsiString msg = "";
switch (message)
	{
	case WM_SETCURSOR:			return msg;
	case WM_NCHITTEST:			return msg;
    case WM_MOUSEMOVE:			return msg;
	case WM_NCMOUSEMOVE:		return msg;
    case WM_PAINT:				msg = "WM_PAINT";				break;
    case WM_ERASEBKGND:			msg = "WM_ERASEBKGND";			break;
    case WM_DRAWITEM:			msg = "WM_DRAWITEM"; 			break;
    case WM_MEASUREITEM:		msg = "WM_MEASUREITEM";			break;
    case WM_VSCROLL:			msg = "WM_VSCROLL";				break;
    case WM_HSCROLL:    		msg = "WM_HSCROLL";				break;
	case WM_COMMAND:			msg = "WM_COMMAND";				break;
	case WM_GETTEXT:			msg = "WM_GETTEXT";				break;
	case WM_GETTEXTLENGTH:      msg = "WM_GETTEXTLENGTH";		break;
	case WM_CTLCOLORMSGBOX:		msg = "WM_CTLCOLORMSGBOX";		break;
	case WM_CTLCOLOREDIT:		msg = "WM_CTLCOLOREDIT";		break;
	case WM_CTLCOLORLISTBOX:	msg = "WM_CTLCOLORLISTBOX";		break;
	case WM_CTLCOLORBTN:		msg = "WM_CTLCOLORBTN";			break;
	case WM_CTLCOLORDLG:		msg = "WM_CTLCOLORDLG";			break;
	case WM_CTLCOLORSCROLLBAR:	msg = "WM_CTLCOLORSCROLLBAR";	break;
	case WM_CTLCOLORSTATIC:		msg = "WM_CTLCOLORSTATIC";		break;
	case WM_CAPTURECHANGED:		msg = "WM_CAPTURECHANGED";		break;
	case WM_MENUSELECT:			msg = "WM_MENUSELECT";			break;
	case WM_SYSCOMMAND:			msg = "WM_SYSCOMMAND";			break;
	case WM_NCCALCSIZE:			msg = "WM_NCCALCSIZE";			break;
	case WM_NCLBUTTONDOWN:		msg = "WM_NCLBUTTONDOWN";		break;
	case WM_NCLBUTTONUP:		msg = "WM_NCLBUTTONUP";			break;
	case WM_GETMINMAXINFO:		msg = "WM_GETMINMAXINFO";		break;
	case WM_MOVE:				msg = "WM_MOVE";				break;
	case WM_SIZE:				msg = "WM_SIZE";				break;
	case WM_WINDOWPOSCHANGING:	msg = "WM_WINDOWPOSCHANGING";	break;
	case WM_WINDOWPOSCHANGED:	msg = "WM_WINDOWPOSCHANGED";	break;
	case WM_SYSKEYUP:			msg = "WM_SYSKEYUP";			break;
	case WM_SYSKEYDOWN:			msg = "WM_SYSKEYDOWN";			break;
	case WM_PARENTNOTIFY:		msg = "WM_PARENTNOTIFY";		break;
	case WM_MOUSEACTIVATE:		msg = "WM_MOUSEACTIVATE";		break;
	case WM_CTLCOLORMSGBOX-1:	msg = "WM_CTLCOLOR?";			break;
	case WM_SETFOCUS:			msg = "WM_SETFOCUS";			break;
	case WM_KILLFOCUS:			msg = "WM_KILLFOCUS";			break;
	case WM_KEYUP:				msg = "WM_KEYUP";				break;
	case WM_NCACTIVATE:         msg = "WM_NCACTIVATE";			break;
	case WM_NCPAINT:			msg = "WM_NCPAINT";				break;
    case WM_ENTERSIZEMOVE:		msg = "WM_ENTERSIZEMOVE";		break;
    case WM_SIZING:				msg = "WM_SIZING";				break;
	case CM_CURSORCHANGED:		msg = "CM_CURSORCHANGED";		break;
	case CM_FONTCHANGED:		msg = "CM_FONTCHANGED";			break;
	case CM_ICONCHANGED:		msg = "CM_ICONCHANGED";			break;
	case CM_ENABLEDCHANGED:		msg = "CM_ENABLEDCHANGED";		break;
	case CM_CTL3DCHANGED:		msg = "CM_CTL3DCHANGED";		break;
	case CM_DIALOGKEY:			msg = "CM_DIALOGKEY";			break;
	case CM_DIALOGCHAR:			msg = "CM_DIALOGCHAR";			break;
	case CM_CHILDKEY:			msg = "CM_CHILDKEY";			break;
	case CM_FOCUSCHANGED:		msg = "CM_FOCUSCHANGED";		break;
	case CM_PARENTSHOWHINTCHANGED:	msg = "CM_PARENTSHOWHINTCHANGED";	break;
	case CN_DRAWITEM:			msg = "CN_DRAWITEM";			break;
	case CN_CTLCOLORBTN:		msg = "CN_CTLCOLORBTN";			break;
	case CN_CTLCOLORDLG:		msg = "CN_CTLCOLORDLG";			break;
	case CN_CTLCOLOREDIT:		msg = "CN_CTLCOLOREDIT";		break;
	case CN_CTLCOLORLISTBOX:	msg = "CN_CTLCOLORLISTBOX";		break;
	case CN_CTLCOLORMSGBOX:		msg = "CN_CTLCOLORMSGBOX";		break;
	case CN_CTLCOLORSCROLLBAR:	msg = "CN_CTLCOLORSCROLLBAR";	break;
	case CN_CTLCOLORSTATIC:		msg = "CN_CTLCOLORSTATIC";		break;
	case CN_SYSKEYDOWN:			msg = "CN_SYSKEYDOWN";			break;
	case CN_NOTIFY:				msg = "CN_NOTIFY";				break;
	case CN_SYSCHAR:			msg = "CN_SYSCHAR";				break;
	case CN_KEYDOWN:			msg = "CN_KEYDOWN";				break;
    default:					msg = "UNKNOWN MESSAGE: " + IntToStr( message );
    }
msg = msg + "  WParam="+IntToHex(HIWORD(wparam),4)+":"+IntToHex(LOWORD(wparam),4);
msg = msg + "  LParam="+IntToHex(HIWORD(lparam),4)+":"+IntToHex(LOWORD(lparam),4);
return msg;
}
//---------------------------------------------------------------------------

