//---------------------------------------------------------------------------
//
// TDirectDrawForm - Flicker free form for Borland C++ Builder & DirectX
// ---------------
//
// (c) 1997 by Darren John Dwyer - All Rights Reserved
//
//
// Disclaimer:
//
// All source code, object files, documentation files, etc. included as
// part this package are supplied subject to the rules in the attached file
// LICENCE.DOC
//
// Please read the file LICENCE.DOC now.
//
//---------------------------------------------------------------------------
#ifndef TDirectDrawFormH
#define TDirectDrawFormH
//---------------------------------------------------------------------------
#include <vcl\Classes.hpp>
#include <vcl\Controls.hpp>
#include <vcl\StdCtrls.hpp>
#include <vcl\Forms.hpp>
//---------------------------------------------------------------------------
#include <vcl\ExtCtrls.hpp>
#include <vcl\Buttons.hpp>
#include <vcl\ComCtrls.hpp>
#include <vcl\Grids.hpp>
#include <vcl\MPlayer.hpp>
#include <vcl\Menus.hpp>
//---------------------------------------------------------------------------
#include "TDirectDraw.h"
//---------------------------------------------------------------------------
class TDirectDrawForm : public TForm
{
    //----------------------------------------------------------------------------------
	// General Notes:
    //
    // (1)	This object "TDirectDrawForm" is designed to be used with the "TDirectDraw"
    //		object written by the same author and will not work without "TDirectDraw".
    //
    // (2)	A TDirectDrawForm is meant to behave exactly the same as the standard TForm
    //		in the manner of dropping components onto the form, etc. except that the
    //		TDirectDrawForm works only with DirectX exclusive mode (ie. full-screen)
	//
    //----------------------------------------------------------------------------------

__published:

public:

    //----------------------------------------------------------------------------------
    // Notes:	Should not need to overload these methods
    //----------------------------------------------------------------------------------

					__fastcall 	TDirectDrawForm(TComponent* Owner);
	virtual			__fastcall 	~TDirectDrawForm();
    virtual	void	__fastcall	Repaint();

    //----------------------------------------------------------------------------------
    // Notes: At the moment, forms are still moveable, attempting to figure way around
    //		  this. For this release: BeforeMove(), AfterMove() and Moveable do not work
    //----------------------------------------------------------------------------------

	virtual	void	__fastcall	BeforeMove(TObject* Sender);
    virtual	void	__fastcall	AfterMove(TObject* Sender);

	__property	bool			Moveable = {read=fFormIsMoveable,write=fFormIsMoveable,default=true};
    __property	bool			Transparent = {read=fTransparent,write=fTransparent,default=false};

    //----------------------------------------------------------------------------------
	// Notes: FramesFlipped is a counter that you can read/write and which is incremented
    //			each time a section of the screen has been updated.
    //----------------------------------------------------------------------------------

    __property	unsigned long	FramesFlipped = {read=fFramesFlipped,write=fFramesFlipped,default=0};

protected:

    //----------------------------------------------------------------------------------
    // Notes:	Internal vars & methods, etc.
    //----------------------------------------------------------------------------------

	void 		__fastcall	WndProc(Messages::TMessage &Message);
	void 		__fastcall	PaintHandler(Messages::TWMPaint &Message);
	void		__fastcall	PaintWindow(HDC hdc);
	void		__fastcall	PaintControls(HDC hdc, TControl* First);

private:

    //----------------------------------------------------------------------------------
    // Notes:	Internal vars & methods, etc.
    //----------------------------------------------------------------------------------

    bool				fHandlingWMPAINT;
    bool				fMicrosoftPlus;
    int					fMouseX;
    int					fMouseY;
    bool				fFormIsActive;
    bool				fFormIsMoving;
    bool				fFormIsMoveable;
    bool				fFormIsBeingRestored;
    bool				fFormIsBeingMinimized;
    bool				fFormIsBeingMaximized;
    bool				fMainFormIsMinimized;
    bool				fMenuIsDisplayed;
    int					fMinimizedWidth;
	int					fBorderX;
	int					fBorderY;
	int					fCaptionHeight;
    int					fButtonWidth;
    int					fButtonHeight;
    int					fScrollbarWidth;
    int					fScrollbarHeight;
    int					fThumbnailHeight;
    int					fIconWidth;
    int					fIconHeight;
    bool				fExcludeMenuYes;
    TRect				fExcludeMenu;
    System::AnsiString	fCaptionFontName;
    System::AnsiString	fSmCaptionFontName;
    unsigned long		fFramesFlipped;
    Graphics::TBitmap*	fUpArrow;
    Graphics::TBitmap*	fDnArrow;
    Graphics::TBitmap*	fLtArrow;
    Graphics::TBitmap*	fRtArrow;
    TImageList*			fSmallIcon;
    bool				fTransparent;
    int					fDisplayWidth;
    int					fDisplayHeight;
    int					fDisplayBitsPerPixel;
    bool				fHScrollBarIsVisible;
    bool				fVScrollBarIsVisible;

    RECT		__fastcall	SelectEntireForm( HDC hdc );
    RECT		__fastcall	SelectClientArea( HDC hdc, RECT r );
    void		__fastcall	PaintWindowFrame( HDC hdc, RECT r );
    void		__fastcall	PaintClientArea( HDC hdc, RECT r );
    void 		__fastcall	PaintDialogFrame( RECT r );
    void 		__fastcall	PaintSingleFrame( RECT r );
    void 		__fastcall	PaintSizeableFrame( RECT r );
    void 		__fastcall	PaintSizeToolWinFrame( RECT r );
    void 		__fastcall	PaintToolWindowFrame( RECT r );

	void 		__fastcall	ExcludeOtherForms( HWND Handle, HDC hdc );
	void		__fastcall	ExcludeMenu( HDC hdc, HMENU hmenu );

	void 		__fastcall	Paint3DButtonFrame( int left, int top, int right, int bottom );
	void 		__fastcall 	Paint3DThumbnail( int left, int top, int right, int bottom );

    void		__fastcall	PaintCloseButton( int left, int top );
    void		__fastcall	PaintCommandButton( int left, int top );
    void		__fastcall	PaintRestoreButton( int left, int top );
    void		__fastcall	PaintMinimizeButton( int left, int top );
    void		__fastcall	PaintMaximizeButton( int left, int top );

	void		__fastcall 	PaintScrollBars( TRect client_r );
	void 		__fastcall	PaintVScrollBar( TRect client_r );
	void 		__fastcall	PaintHScrollBar( TRect client_r );

	void		__fastcall	PaintListBox( HDC hdc, TListBox* listbox );
	void 		__fastcall	PaintControlVScrollBar( HDC hdc, HWND hwnd, int left, int top, int width, int height, int pos, int rowoffset, int rowsperpage, int totalrows );
	void 		__fastcall	PaintControlHScrollBar( HDC hdc, HWND hwnd, int left, int top, int width, int height, int pos, int rowoffset, int rowsperpage, int totalrows );

    void		__fastcall	HandleException(TObject* Sender);
    void		__fastcall	OnIdle(TObject* Sender, bool &done);

    System::AnsiString	__fastcall MessageStr( int message, int wparam, int lparam );
};
//---------------------------------------------------------------------------
extern TDirectDrawForm *DirectDrawForm;
//---------------------------------------------------------------------------
#endif


